import dessia_common.utils.diff as dcd
import random
from time import time

empty_sequence = []
one_element_sequence = [1]
two_element_sequence = [1, 2]

print("Building Sequences...\n")

start = time()
n_each = 200000
n_each_empty = 10000
n_each_simple = (n_each - n_each_empty)//2
same_sequences = [empty_sequence]*n_each_empty
same_sequences += [one_element_sequence]*n_each_simple
same_sequences += [two_element_sequence]*n_each_simple
same_length_sequences = [[random.randint(0, n_each) for _ in range(100)] for _ in range(n_each)]
random_length_sequences = [[random.randint(0, n_each) for _ in range(random.randint(0, 100))] for _ in range(n_each)]
sequences = same_sequences + same_length_sequences + random_length_sequences
end = time()

print(f"{len(sequences)} sequences built in {end - start}s.\n\nStarting perf testing...\n")
n_empty = sequences.count(empty_sequence) - n_each_empty
n_one = sequences.count(one_element_sequence) - n_each_simple
n_two = sequences.count(two_element_sequence) - n_each_simple
max_uniques = len(same_length_sequences) + len(random_length_sequences) - 3 + n_empty + n_one + n_two
expected = f" (Expected around {max_uniques})"


start = time()
dcd.sequence_hash(sequences)
end = time()

duration = end - start
print(f"Sequence hash : Lasted {duration}s\n")


hashes = [dcd.sequence_hash(s) for s in sequences]
unique_hashes = len(set(hashes))
total_hashes = len(hashes)
efficiency = unique_hashes/total_hashes
max_efficiency = max_uniques/total_hashes
print(f"Hash efficiency : \n"
      f" - Unique : {unique_hashes}{expected}\n"
      f" - Total : {total_hashes}\n"
      f" - Efficiency : {efficiency} (max around {max_efficiency})\n")

assert unique_hashes >= 0.90*max_uniques
assert duration <= 1e-3


def test_hash_function(hash_func):
    # Empty sequences hashes
    empty_seq1 = []
    empty_seq2 = ()
    assert hash_func(empty_seq1) == hash_func(empty_seq2)

    # Different size hashes
    seq1 = [1, 2, 3]
    seq2 = [1, 2, 3, 4]
    assert hash_func(seq1) != hash_func(seq2)

    # Different types sequences
    seq3 = [1, 2, 'a', 'b']
    seq4 = [1, 2, 'b', 'a']
    assert hash_func(seq3) != hash_func(seq4)

    # Sequences with mutables
    # TODO Add a random seed to not only check first and last, but random ones
    # TODO Not mandatory to have diffrent types
    # seq5 = [1, [2, 3], 4]
    # seq6 = [1, [3, 2], 4]
    # assert hash_func(seq5) != hash_func(seq6)

    # Sequence with different hash computation
    # TODO Check, should probably be equal
    # seq7 = [1, 2.0, 3]
    # seq8 = [1, 2, 3]
    # assert hash_func(seq7) != hash_func(seq8)

    # Random sequences
    seq11 = random.sample(range(100), 10)
    seq12 = random.sample(range(100), 10)
    assert hash_func(seq11) != hash_func(seq12)

test_hash_function(dcd.sequence_hash)


# TODO Check following for test case
# def test_hash_quality_recursive(hash_func, max_depth=5, min_size=2, max_size=10):
#     # Collision test
#     num_keys = 100000
#     collisions = 0
#     hashes = {}
#     for _ in range(num_keys):
#         key = generate_random_recursive_sequence(max_depth, min_size, max_size)
#         h = hash_func(key)
#         if h in hashes:
#             collisions += 1
#         else:
#             hashes[h] = key
#     print("Nombre de collisions:", collisions)
#
#     # Perf tests
#     import time
#     key_depths = [2, 3, 4, 5]
#     for key_depth in key_depths:
#         keys = [generate_random_recursive_sequence(key_depth, min_size, max_size) for _ in range(num_keys)]
#         start_time = time.time()
#         for key in keys:
#             # TODO Unfinished test
#
# def generate_random_recursive_sequence(max_depth, min_size, max_size):
#   if max_depth == 0:
#         return "".join(random.choices(string.ascii_letters, k=random.randint(min_size, max_size)))
#   else:
#         size = random.randint(min_size, max_size)
#         return [generate_random_recursive_sequence(max_depth - 1, min_size, max_size) for _ in
#                 range(size)]

# def test_hash_dispersion(hash_func, num_keys, num_buckets, key_length):
#     # Init a sequence that stores number of keys from each bucket
#     buckets = [0] * num_buckets
#
#     # Random sequences
#     keys = ["".join(random.choices(string.ascii_letters, k=key_length)) for _ in range(num_keys)]
#
#     # Compute hash and add to buckt
#     for key in keys:
#         bucket = hash_func(key) % num_buckets
#         buckets[bucket] += 1
#
#
#     # Compute variance and variation coeff for keys in each bucket
#     mean = num_keys / num_buckets
#     variance = sum((bucket - mean) ** 2 for bucket in buckets) / num_buckets
#     cv = (variance / mean) ** 0.5
#
#     return cv
#
# cv = test_hash_dispersion(hash_func, 100000, 1000, 20)
# print("Coefficient de variation:", cv)
print("script 'hashes.py' has passed")



"""
A module that aims to list all possibilities of data formats offered by Dessia.

It should be used as a repertory of rules and available typing.

Some general rules :

- Lists are homogeneous, IE, they should not contain several types of elements
    ex : List[int], List[str], List[Class], ...
- Tuples can, therefore, be used as heterogeneous sequences, thanks to the fact that they are immutable.
    ex : Tuple[str, int, CustomClass] is a tuple like this :
        t = ('tuple', 1, custom_class_object)
- Dict is used whenever a dynamic structure should be defined. It takes only one possible type for keys
  and one possible type for values.
    ex : Dict[str, bool] is a dict like :
        d = {'key0': True, 'key1': False, 'another_key': False,...}
- As opposed to this, a non-standalone_in_db class defines a static structure,
  with a defined number of given, expected keys & types of their values.

In addition to types & generics (brought by DessiaObject), this module can also be seen as a template for Dessia
coding/naming style & convention.
"""

from math import floor, ceil, cos
from typing import Dict, List, Tuple, Union, Any
import time
from numpy import linspace

try:
    import volmdlr as vm
    from volmdlr import primitives2d as p2d
    from volmdlr import primitives3d as p3d
    import plot_data
    import plot_data.colors
except ImportError as err:
    print("Couldn't import plot_data or volmdlr due to the following exception : ", err)

from dessia_common.core import DessiaObject, PhysicalObject, MovingObject
from dessia_common.typings import InstanceOf
from dessia_common.measures import Distance
from dessia_common.exports import MarkdownWriter

from dessia_common.files import BinaryFile, StringFile

from dessia_common.decorators import plot_data_view, markdown_view, cad_view


class EmbeddedBuiltinsSubobject(PhysicalObject):
    """
    An object that is not standalone and gather builtins types (float, int, bool, str & Distance).

    :param distarg: A Distance with units
    :param floatarg: A float
    :param intarg: An integer
    :param boolarg: A Boolean
    :param name: Object's name
    """

    _standalone_in_db = False

    def __init__(self, distarg: Distance, floatarg: float, intarg: int,
                 boolarg: bool, name: str = "Standalone Subobject"):
        self.distarg = distarg
        self.floatarg = floatarg
        self.intarg = intarg
        self.boolarg = boolarg

        PhysicalObject.__init__(self, name=name)

    @classmethod
    def generate(cls, seed: int) -> 'EmbeddedBuiltinsSubobject':
        """ Generate an embedded sub-object with default values computed from a seed. """
        floatarg = 0.3
        distarg = Distance(1.7 * floatarg * seed)
        intarg = seed
        boolarg = bool(seed % 2)
        name = f"EmbeddedSubobject{seed}"
        return cls(distarg=distarg, floatarg=floatarg, intarg=intarg, boolarg=boolarg, name=name)

    @classmethod
    def generate_many(cls, seed: int) -> List['EmbeddedBuiltinsSubobject']:
        """ Generate many embedded subobjects with default values computed from a seed. """
        return [cls.generate((i + 1) * 1000) for i in range(seed)]

    def contour(self):
        """ Square contour of an embedded sub-object, for testing purpose. """
        origin = self.floatarg
        points = [vm.Point2D(origin, origin), vm.Point2D(origin, origin + 1),
                  vm.Point2D(origin + 1, origin + 1), vm.Point2D(origin + 1, origin)]
        return p2d.ClosedRoundedLineSegments2D(points=points, radius={})

    def plot_data(self, reference_path: str = "#", **kwargs):
        """ Bare text, for testing purpose. """
        primitives = [plot_data.Text(comment="Test with text", position_x=0, position_y=0),
                      plot_data.Text(comment="Test with text", position_x=0, position_y=0)]
        primitives_group = plot_data.PrimitiveGroup(primitives=primitives)
        return [primitives_group]

    def voldmlr_primitives(self):
        """ Volmdlr primitives of the squared contour. """
        contour = self.contour()
        volumes = [p3d.ExtrudedProfile(vm.O3D, vm.X3D, vm.Z3D, contour, [], vm.Y3D)]
        return volumes


class StandaloneBuiltinsSubobject(EmbeddedBuiltinsSubobject):
    """
    Overwrite EmbeddedBuiltinsObject to make it standalone.

    :param distarg: A Distance with units
    :param floatarg: A float
    :param intarg: An integer
    :param boolarg: A boolean
    :param name: Object's name
    """

    _standalone_in_db = True

    def __init__(self, distarg: Distance, floatarg: float, intarg: int,
                 boolarg: bool, name: str = "Standalone Subobject"):

        EmbeddedBuiltinsSubobject.__init__(self, distarg=distarg, floatarg=floatarg,
                                           intarg=intarg, boolarg=boolarg, name=name)

    @property
    def property_method(self):
        """ Property method for low-code testing purpose. """
        return self.floatarg * self.intarg


DEF_EBS = EmbeddedBuiltinsSubobject.generate(1)
DEF_SBS = StandaloneBuiltinsSubobject.generate(1)


class EnhancedStandaloneSubobject(StandaloneBuiltinsSubobject):
    """ Overwrite StandaloneSubobject, principally for InstanceOf and Union typing testing purpose. """

    def __init__(self, floatarg: Distance, name: str = "Standalone Subobject"):
        StandaloneBuiltinsSubobject.__init__(self, distarg=floatarg, floatarg=floatarg, intarg=floor(floatarg),
                                             boolarg=floatarg.is_integer(), name=name)

    @classmethod
    def generate(cls, seed: int) -> 'EnhancedStandaloneSubobject':
        """ Generate an enhanced sub-object with default values computed from a seed. """
        floatarg = Distance(1.2 * seed)
        name = f"EnhancedStandaloneSubobject{seed}"
        return cls(floatarg=floatarg, name=name)


DEF_ESS = EnhancedStandaloneSubobject.generate(1)


class InheritingStandaloneSubobject(StandaloneBuiltinsSubobject):
    """ Overwrite StandaloneSubobject, principally for InstanceOf and Union typing testing purpose. """

    def __init__(self, distarg: Distance, floatarg: float, intarg: int, boolarg: bool,
                 name: str = "Inheriting Standalone Subobject"):
        self.strarg = name

        StandaloneBuiltinsSubobject.__init__(self, distarg=distarg, floatarg=floatarg, intarg=intarg,
                                             boolarg=boolarg, name=name)

    @classmethod
    def generate(cls, seed: int) -> 'InheritingStandaloneSubobject':
        """ Generate an inheriting sub-object with default values computed from a seed. """
        distarg = Distance(0.7 * seed)
        floatarg = 0.1 * seed
        intarg = seed * 3
        boolarg = bool(intarg % 2)
        name = "Inheriting Standalone Subobject" + str(seed)
        return cls(floatarg=floatarg, distarg=distarg, intarg=seed * 3, boolarg=boolarg, name=name)


DEF_ISS = InheritingStandaloneSubobject.generate(1)


class EmbeddedSubobject(DessiaObject):
    """ EmbeddedSubobject, in order to test non standalone features. """

    def __init__(self, embedded_list: List[int] = None, name: str = "Embedded Subobject"):
        if embedded_list is None:
            self.embedded_list = [1, 2, 3]
        else:
            self.embedded_list = embedded_list

        DessiaObject.__init__(self, name=name)

    @classmethod
    def generate(cls, seed: int) -> 'EmbeddedSubobject':
        """ Generate an embedded sub-object with default values computed from a seed. """
        if not bool(seed % 2):
            embedded_list = list(range(int(seed / 2)))
        else:
            embedded_list = None
        name = "Embedded Subobject" + str(seed)
        return cls(embedded_list=embedded_list, name=name)

    @classmethod
    def generate_many(cls, seed: int) -> List['EmbeddedSubobject']:
        """ Generate many embedded subobjects with default values computed from a seed. """
        return [cls.generate(i) for i in range(ceil(seed / 3))]


class EnhancedEmbeddedSubobject(EmbeddedSubobject):
    """ Overwrite EmbeddedSubobject, principally for InstanceOf and Union typing testing purpose. """

    def __init__(self, embedded_list: List[int] = None, embedded_array: List[List[float]] = None,
                 name: str = "Enhanced Embedded Subobject"):
        self.embedded_array = embedded_array

        EmbeddedSubobject.__init__(self, embedded_list=embedded_list, name=name)

    @classmethod
    def generate(cls, seed: int) -> 'EnhancedEmbeddedSubobject':
        """ Generate an embedded sub-object with default values computed from a seed. """
        embedded_list = [seed]
        embedded_array = [[seed, seed * 10, seed * 10]] * seed
        name = f"Embedded Subobject{seed}"
        return cls(embedded_list=embedded_list, embedded_array=embedded_array, name=name)


DEF_ES = EmbeddedSubobject.generate(10)
DEF_EES = EnhancedEmbeddedSubobject.generate(3)


UnionArg = Union[EmbeddedSubobject, EnhancedEmbeddedSubobject]


class StandaloneObject(MovingObject):
    """
    Standalone Object for testing purpose.

    :param standalone_subobject: A sub-object that is standalone_in_db, for development purpose
    :param embedded_subobject: A sub-object that isn't standalone_in_db, for development purpose
    :param dynamic_dict: A variable length dict
    :param tuple_arg: A heterogeneous sequence
    """

    _standalone_in_db = True
    _generic_eq = True
    _allowed_methods = ["add_standalone_object", "count_until", "add_float",
                        "generate_from_text", "generate_from_bin", "method_without_arg", "ill_defined_method"]
    _non_serializable_attributes = ["samples"]

    def __init__(self, standalone_subobject: StandaloneBuiltinsSubobject, embedded_subobject: EmbeddedSubobject,
                 dynamic_dict: Dict[str, bool], float_dict: Dict[str, float], string_dict: Dict[str, str],
                 tuple_arg: Tuple[str, int], object_list: List[StandaloneBuiltinsSubobject],
                 subobject_list: List[EmbeddedSubobject], builtin_list: List[int], union_arg: List[UnionArg],
                 subclass_arg: InstanceOf[StandaloneBuiltinsSubobject], array_arg: List[List[float]],
                 name: str = "Standalone Object Demo"):
        self.union_arg = union_arg
        self.builtin_list = builtin_list
        self.subobject_list = subobject_list
        self.object_list = object_list
        self.tuple_arg = tuple_arg
        self.dynamic_dict = dynamic_dict
        self.float_dict = float_dict
        self.string_dict = string_dict
        self.standalone_subobject = standalone_subobject
        self.embedded_subobject = embedded_subobject
        self.subclass_arg = subclass_arg
        self.array_arg = array_arg

        self._samples = None

        MovingObject.__init__(self, name=name)

    @classmethod
    def generate(cls, seed: int, name: str = "Standalone Object Demo") -> 'StandaloneObject':
        """ Generate an object with default values computed from a seed. """
        dynamic_dict = {f"n{i}": bool(seed % 2) for i in range(seed)}
        float_dict = {f"k{i}": seed * 1.09 for i in range(seed)}
        string_dict = {f"key{i}": f"value{i}" for i in range(seed)}
        builtin_list = [seed] * seed
        array_arg = [builtin_list] * 3
        union_arg = [EnhancedEmbeddedSubobject.generate(seed), EmbeddedSubobject.generate(seed)]
        if not bool(seed % 2):
            subclass_arg = StandaloneBuiltinsSubobject.generate(-seed)
        else:
            subclass_arg = InheritingStandaloneSubobject.generate(seed)
        return cls(standalone_subobject=StandaloneBuiltinsSubobject.generate(seed),
                   embedded_subobject=EmbeddedSubobject.generate(seed),
                   dynamic_dict=dynamic_dict, float_dict=float_dict, string_dict=string_dict,
                   tuple_arg=("value", seed * 3), object_list=StandaloneBuiltinsSubobject.generate_many(seed),
                   subobject_list=EmbeddedSubobject.generate_many(seed), builtin_list=builtin_list, union_arg=union_arg,
                   subclass_arg=subclass_arg, array_arg=array_arg, name=name)

    @classmethod
    def generate_with_many_subobjects(cls, seed: int, number: int, name: str = "Populated SO"):
        """ Generate an object with a lot of subobjects. """
        standalone_object = cls.generate(seed=seed, name=name)
        subobjects = StandaloneBuiltinsSubobject.generate_many(seed=seed * number)
        standalone_object.subclass_arg = subobjects
        return standalone_object

    @classmethod
    def generate_from_bin(cls, stream: BinaryFile):
        """ Generate an object from bin file in order to test frontend forms and backend streams. """
        # User need to decode the binary as he sees fit
        my_string = stream.read().decode("utf8")
        my_file_name = stream.filename
        _, raw_seed = my_string.split(",")
        seed = int(raw_seed.strip())
        return cls.generate(seed=seed, name=my_file_name)

    @classmethod
    def generate_from_text(cls, stream: StringFile):
        """ Generate an object from text file in order to test frontend forms and backend streams. """
        my_text = stream.getvalue()
        my_file_name = stream.filename
        _, raw_seed = my_text.split(",")
        seed = int(raw_seed.strip())
        return cls.generate(seed=seed, name=my_file_name)

    def add_standalone_object(self, object_: StandaloneBuiltinsSubobject):
        """
        Add a standalone object to object_list.

        It doesn't return anything, hence, API will update object when computing from frontend.
        """
        self.object_list.append(object_)

    def add_float(self, value: float = 1) -> StandaloneBuiltinsSubobject:
        """
        Add value to its standalone sub-object floatarg property and return it.

        API should replace standalone_subobject as it is returned
        """
        self.standalone_subobject.floatarg += value
        return self.standalone_subobject

    def append_union_arg(self, object_: UnionArg):
        """ Append an object with corresponding type to union_arg, for testing purpose. """
        self.union_arg.append(object_)

    def method_without_arg(self):
        """ It should be OK and be in method_schema. """
        return self

    def contour(self):
        """ Squared contour. """
        intarg = self.standalone_subobject.intarg
        points = [vm.Point2D(intarg, intarg), vm.Point2D(intarg, intarg + 1),
                  vm.Point2D(intarg + 1, intarg + 1), vm.Point2D(intarg + 1, 0)]

        crls = p2d.ClosedRoundedLineSegments2D(points=points, radius={})
        return crls

    def volmdlr_primitives(self, **kwargs):
        """ Volmdlr primitives of a cube. """
        subcube = self.standalone_subobject.voldmlr_primitives()[0]
        contour = self.contour()
        cube = p3d.ExtrudedProfile(plane_origin=vm.Point3D(0, 1, -1), x=vm.X3D, y=vm.Z3D,
                                   outer_contour2d=contour, inner_contours2d=[], extrusion_vector=vm.Y3D)
        return [subcube, cube]

    def volmdlr_primitives_step_frames(self):
        """ Steps to show the cube moving on frontend. """
        frame0 = vm.Frame3D(vm.O3D.copy(), vm.X3D.copy(), vm.Y3D.copy(), vm.Z3D.copy())
        frame11 = frame0.rotation(center=vm.O3D, axis=vm.Y3D, angle=0.7)
        frame21 = frame11.translation(offset=vm.Y3D)
        frame31 = frame21.rotation(center=vm.O3D, axis=vm.Y3D, angle=0.7)

        frame12 = frame0.translation(offset=vm.Z3D)
        frame22 = frame12.translation(offset=vm.X3D)
        frame32 = frame22.translation(offset=vm.X3D)
        return [[frame0, frame0], [frame11, frame12], [frame21, frame22], [frame31, frame32]]

    @cad_view("CAD With Selector")
    def display_cad(self):
        """ Volmdlr primitives of 'cubes'. """
        subcube = self.standalone_subobject.voldmlr_primitives()[0]
        contour = self.contour()
        cube = p3d.ExtrudedProfile(plane_origin=vm.Point3D(0, 1, -1), x=vm.X3D, y=vm.Z3D,
                                   outer_contour2d=contour, inner_contours2d=[], extrusion_vector=vm.Y3D)
        primitives = [subcube, cube]
        return vm.core.MovingVolumeModel(primitives, self.volmdlr_primitives_step_frames()).babylon_data()

    @plot_data_view("2D View")
    def primitives(self):
        """ Test plot data decorator for primitives. """
        contour = self.standalone_subobject.contour().plot_data()
        return plot_data.PrimitiveGroup(primitives=[contour], name="Contour")

    @plot_data_view("Scatter Plot")
    def scatter_plot(self, **kwargs):
        """ Test plot data decorator for scatter plots. """
        reference_path = kwargs.get("reference_path", "#")
        attributes = ["cx", "cy"]
        tooltip = plot_data.Tooltip(attributes=attributes, name="Tooltips")
        samples = self._random_samples(reference_path)
        return plot_data.Scatter(elements=samples, tooltip=tooltip, x_variable=attributes[0],
                                 y_variable=attributes[1], name="Scatter Plot")

    @plot_data_view("Parallel Plot")
    def parallel_plot(self, reference_path: str = "#"):
        """ Test plot data decorator for parallel plots. """
        samples = self._random_samples(reference_path)
        return plot_data.ParallelPlot(elements=samples, axes=["cx", "cy", "label"], name="Parallel Plot")

    @plot_data_view("Multiplot", load_by_default=True)
    def multiplot(self, reference_path: str = "#"):
        """ Test plot data decorator for multiple plots. """
        scatter_plot = self.scatter_plot(reference_path=reference_path)
        parallel_plot = self.parallel_plot(reference_path=reference_path)
        objects = [scatter_plot, parallel_plot]
        sizes = [plot_data.Window(width=560, height=300), plot_data.Window(width=560, height=300)]
        samples = self._random_samples(reference_path=reference_path)
        return plot_data.MultiplePlots(elements=samples, plots=objects, sizes=sizes,
                                       coords=[(0, 0), (300, 0)], name="Multiple Plot")

    def _random_samples(self, reference_path: str = "#"):
        """ A dummy method to generate plot_data Samples for testing purpose. """
        if self._samples is None:
            samples = []
            for i in range(500):
                abscissa = round((i % 17) / 2.3, 3)
                values = {"cx": abscissa, "cy": 100 + abscissa * i / 100, "label": f"label{i % 5}"}
                samples.append(plot_data.Sample(values=values, reference_path=f"{reference_path}/path/to/element/{i}",
                                                name=f"Point{i}"))
            self._samples = samples
        return self._samples

    def ill_defined_method(self, arg0, arg1=1, arg2: int = 10, arg3=3):
        """ Define a docstring for testing parsing purpose. """
        nok_string = "This is a bad coding behavior"
        ok_string = "This could be OK as temporary attr"
        arg0, arg1, arg2 = (arg1, arg2, arg0)
        self.maldefined_attr = nok_string
        self._ok_attribute = ok_string

        computation = f"{nok_string} or {ok_string}"

        return computation

    @markdown_view(selector="Markdown", load_by_default=True)
    def to_markdown(self):
        """ Write a standard markdown of StandaloneObject. """
        contents = """
        # Quem Stygios dumque

        ## Recursus erat aere decus Lemnicolae

        Lorem markdownum laetum senior quod Libys utroque
        *mirantibus teneat aevo*, aquis.
        Procumbit eandem ensis, erigor intercepta, quae habitabat nostro
        *et hoc que* enim: inpulit.
        Mecum ferat **fecissem** vale per myricae suis
        quas turba potentior mentita.
        Annis nunc, picae erat quis minatur dare Diana redimitus
        [Clymene venisses sinat](http://est.net/umbram.html)
        protinus pulchra, sucos! Tanta haec varios tuaque,
        nisi Erigonen si aquae Hippomene inguine murmur.

        1. Poma enim dextra icta capillis extinctum foedera
        2. Mediis requirit exercita ascendere fecisse sola
        3. Sua externis tigride saevarum
        4. Aves est pendebant sume latentis

        ## Suum videre quondam generis dolentem simul femineos

        Ille lacus progenitore Cycnum pressa, excidit silva
        [crudus](http://www.domino.com/nequevox), boum ducem vocari,
        ne monte tanto harenae.
        Opus Aesone excipit adempto.
        Inpius illa latratu atque sed praedam,
        ille construit intravit concipit,
        concha dedit, qua audit calathosque.
        Dedit putrefacta cortex.
        Tenet aut carmina quod proditione media; pro ense medicina
        vita repetit adrectisque inops e sentiat.

        > Imagine caesaries superbos muneraque *ne terras* cunctis.
        Diversae Hesioneque
        > numinis regia at anima nascuntur Iovis.
        Sua fama quoque capillos lugubris
        > **egimus**, a ingenti [Ericthonio](http://raptos.org/lucem)
        iubebat!

        ## Ponderis venit veteris mihi tofis

        Propensum discedunt, iacere dedisti; lens potest caelo,
        felix flamma caecus decet excipit.
        *Aurum occiderat*, retro cum, quorum *Diana timuere At*.
        Ait Labros hasta mundi, **ut est** ruit nosse o gravet!

        ## Qui aether undis nulla

        Homines oppidaque nominibus devexo genitoris quoque,
        praesensque rota Saturnia.
        Auras cecinit fera quae mirantum imbris,
        Gratia verba incesto, sed visa contigit
        saepe adicit trepidant.
        [Siqua radiis quod](http://www.naris-pectebant.org/comeset)
        ad duabus alienisque, sponte; dumb.

        Occidit Babylonia dubitare. Vultus cui: erat dea!
        Iam ense forma est se, tibi pedem adfectat nec nostra.
        Armenta socium nutrix [precatur](http://in-fraxinus.io/)
        aderam, quam mentem Elin labor auctor
        potentia prodidit inmitibus duo di?
        Verum a, tuo quoque nec Mysum per posses;
        vigor danda meruit: tecum audire responsa
        [conplexae](http://quis.io/disrestat.html) et alios.

        Agros grata illo animo mei nova, in magis furens et
        [modo](http://pondere.com/aquis) dimittere ubi neque es!
        Sua qua ac ire una facit Alcmene coepere
        arduus quae vestigia aliquis; meritorum Dorylas, scindunt.
        """
        contents += "\n## Attribute Table\n\n"
        contents += MarkdownWriter(print_limit=25, table_limit=None).object_table(self)
        return contents
    
    @plot_data_view("Graph 2D")
    def graph(self):
        """
        Base plot_data method. Overwrite this to display 2D or graphs on platform.

        Should return a list of plot_data's objects.
        """
        attributes = ["timestep", "electric current"]
        tooltip = plot_data.Tooltip(attributes=attributes)
        timesteps = linspace(0, 20, 20)
        current1 = [t ** 2 for t in timesteps]
        elements1 = [plot_data.Sample({"timestep": t, "electric current": c}) for t, c in zip(timesteps, current1)]

        # The previous line instantiates a dataset with limited arguments but some customization is available
        point_style = plot_data.PointStyle(color_fill=plot_data.colors.RED, color_stroke=plot_data.colors.BLACK)
        edge_style = plot_data.EdgeStyle(color_stroke=plot_data.colors.BLUE, dashline=[10, 5])

        custom_dataset = plot_data.Dataset(elements=elements1, name="I = f(t)", tooltip=tooltip,
                                           point_style=point_style, edge_style=edge_style)

        # Now let's create another dataset for the purpose of this exercise
        timesteps = linspace(0, 20, 100)
        current2 = [100 * (1 + cos(t)) for t in timesteps]
        elements2 = [plot_data.Sample({"timestep": t, "electric current": c}) for t, c in zip(timesteps, current2)]

        dataset2 = plot_data.Dataset(elements=elements2, name="I2 = f(t)")
        return plot_data.Graph2D(graphs=[custom_dataset, dataset2], x_variable=attributes[0], y_variable=attributes[1])

    def count_until(self, duration: float, raise_error: bool = False):
        """
        Test long execution with a user-input duration.

        :param duration: Duration of the method in s
        :param raise_error: Whether the computation should raise an error or not at the end
        """
        starting_time = time.time()
        current_time = time.time()
        last_duration = round(current_time - starting_time)
        while current_time - starting_time <= duration:
            current_time = time.time()
            current_duration = current_time - starting_time
            if current_duration > last_duration + 1:
                last_duration = round(current_time - starting_time)
                print(round(current_duration))

        if raise_error:
            raise RuntimeError(f"Evaluation stopped after {duration}s")


DEF_SO = StandaloneObject.generate(1)


class StandaloneObjectWithDefaultValues(StandaloneObject):
    """ Overwrite StandaloneObject to set default values to it. For frontend forms testing purpose. """

    _non_editable_attributes = ["intarg", "strarg"]

    def __init__(self, standalone_subobject: StandaloneBuiltinsSubobject = DEF_SBS,
                 embedded_subobject: EmbeddedSubobject = DEF_ES,
                 dynamic_dict: Dict[str, bool] = None,
                 float_dict: Dict[str, float] = None,
                 string_dict: Dict[str, str] = None,
                 tuple_arg: Tuple[str, int] = ("Default Tuple", 0),
                 object_list: List[StandaloneBuiltinsSubobject] = None,
                 subobject_list: List[EmbeddedSubobject] = None,
                 builtin_list: List[int] = None,
                 union_arg: List[UnionArg] = None,
                 subclass_arg: InstanceOf[StandaloneBuiltinsSubobject] = DEF_ISS,
                 array_arg: List[List[float]] = None,
                 name: str = "Standalone Object Demo"):
        if dynamic_dict is None:
            dynamic_dict = {}
        if float_dict is None:
            float_dict = {}
        if string_dict is None:
            string_dict = {}
        if object_list is None:
            object_list = [DEF_SBS]
        if subobject_list is None:
            subobject_list = [DEF_ES]
        if builtin_list is None:
            builtin_list = [1, 2, 3, 4, 5]
        if union_arg is None:
            union_arg = [DEF_EES, DEF_ES]
        if array_arg is None:
            array_arg = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]

        StandaloneObject.__init__(self, standalone_subobject=standalone_subobject,
                                  embedded_subobject=embedded_subobject, dynamic_dict=dynamic_dict,
                                  float_dict=float_dict, string_dict=string_dict, tuple_arg=tuple_arg,
                                  object_list=object_list, subobject_list=subobject_list, builtin_list=builtin_list,
                                  union_arg=union_arg, subclass_arg=subclass_arg, array_arg=array_arg, name=name)


DEF_SOWDV = StandaloneObjectWithDefaultValues()


class ObjectWithOtherTypings(DessiaObject):
    """ Dummy class to test some typing schemas. """

    def __init__(self, undefined_type_attribute: Any, name: str = ""):
        self.undefined_type_attribute = undefined_type_attribute

        DessiaObject.__init__(self, name=name)


class MovingStandaloneObject(MovingObject):
    """ Overwrite StandaloneObject to make its 3D move. """

    _standalone_in_db = True

    def __init__(self, origin: float, name: str = ""):
        self.origin = origin
        MovingObject.__init__(self, name=name)

    @classmethod
    def generate(cls, seed: int):
        """ Generate an object with default values from seed. """
        return cls(origin=1.3 * seed, name=f"moving_{seed}")

    def contour(self):
        """ Squared contour. """
        points = [vm.Point2D(self.origin, self.origin), vm.Point2D(self.origin, self.origin + 1),
                  vm.Point2D(self.origin + 1, self.origin + 1), vm.Point2D(self.origin + 1, 0)]

        crls = p2d.ClosedRoundedLineSegments2D(points=points, radius={})
        return crls

    def volmdlr_primitives(self, **kwargs):
        """ A cube. """
        contour = self.contour()
        volume = p3d.ExtrudedProfile(plane_origin=vm.O3D, x=vm.X3D, y=vm.Z3D, outer_contour2d=contour,
                                     inner_contours2d=[], extrusion_vector=vm.Y3D)
        return [volume]

    def volmdlr_primitives_step_frames(self):
        """ A moving cube. """
        frame1 = vm.Frame3D(vm.O3D.copy(), vm.X3D.copy(), vm.Y3D.copy(), vm.Z3D.copy())
        frame2 = frame1.rotation(center=vm.O3D, axis=vm.Y3D, angle=0.7)
        frame3 = frame2.translation(offset=vm.Y3D)
        frame4 = frame3.rotation(center=vm.O3D, axis=vm.Y3D, angle=0.7)
        return [[frame1], [frame2], [frame3], [frame4]]


class Generator(DessiaObject):
    """
    A class that allow to generate several StandaloneObjects from different parameters.

    :param parameter: An "offset" for the seed that will be used in generation
    :param nb_solutions: The max number of solutions that will be generated
    :param name: The name of the Generator. It is not used in object generation
    """

    _standalone_in_db = True

    def __init__(self, parameter: int, nb_solutions: int = 25, models: List[StandaloneObject] = None, name: str = ""):
        self.parameter = parameter
        self.nb_solutions = nb_solutions
        self.models = models

        DessiaObject.__init__(self, name=name)

    def generate(self) -> List[StandaloneObject]:
        """ Generate a list of Standalone objects in order to emulate bots' generators. """
        self.models = [StandaloneObject.generate(self.parameter + i) for i in range(self.nb_solutions)]
        return self.models


class Optimizer(DessiaObject):
    """
    Mock an optimization process. Emulates bots Optimizers.

    :param model_to_optimize: An object which will be modified (one of its attributes)
    :param name: Name of the optimizer. Will not be used in the optimization process
    """

    _standalone_in_db = True

    def __init__(self, model_to_optimize: StandaloneObject, name: str = ''):
        self.model_to_optimize = model_to_optimize

        DessiaObject.__init__(self, name=name)

    def optimize(self, optimization_value: int = 3) -> int:
        """
        Sum model value with given one.

        :param optimization_value: value that will be added to model's intarg attribute
        """
        self.model_to_optimize.standalone_subobject.intarg += optimization_value
        return self.model_to_optimize.standalone_subobject.intarg


class Container(DessiaObject):
    """ Gather a list of Standalone objects as a container. For 'Catalog' behavior testing purpose. """

    _standalone_in_db = True
    _allowed_methods = ["generate_from_text_files"]

    def __init__(self, models: List[StandaloneObject] = None, name: str = ""):
        if models is None:
            self.models = []
        else:
            self.models = models

        DessiaObject.__init__(self, name=name)

    @classmethod
    def generate_from_text_files(cls, files: List[StringFile], name: str = "Generated from text files"):
        """ Generate catalog from several text files. """
        models = [StandaloneObject.generate_from_text(file) for file in files]
        return cls(models=models, name=name)


class NotStandalone(DessiaObject):
    """ A simple non-standalone class. """

    def __init__(self, attribute: int, name: str = ""):
        self.attribute = attribute
        DessiaObject.__init__(self, name=name)


class BottomLevel(DessiaObject):
    """ A simple class at the bottom of the data structure. """

    _standalone_in_db = True

    def __init__(self, attributes: List[NotStandalone] = None, name: str = ""):
        if attributes is None:
            self.attributes = []
        else:
            self.attributes = attributes

        DessiaObject.__init__(self, name=name)


class MidLevel(DessiaObject):
    """ A simple class at the mid level of the data structure. """

    _standalone_in_db = True
    _allowed_methods = ["generate_with_references"]

    def __init__(self, bottom_level: BottomLevel = None, name: str = ""):
        self.bottom_level = bottom_level

        DessiaObject.__init__(self, name=name)

    @classmethod
    def generate_with_references(cls, name: str = "Result Name"):
        """ A fake class generator. """
        object1 = NotStandalone(attribute=1, name="1")
        object2 = NotStandalone(attribute=1, name="2")
        bottom_level = BottomLevel([object1, object2])
        return cls(bottom_level=bottom_level, name=name)



""" Distances operators for metrics on Datasets. """

import numpy as npy
from scipy.spatial.distance import mahalanobis


def diff_list(list_a, list_b):
    """
    Difference between to lists.

    :param list_a: First list
    :type list_a: List[float]

    :param list_b: Second list
    :type list_b: List[float]

    :return: a generator of the difference between each element
    :rtype: generator
    """
    return (a - b for a, b in zip(list_a, list_b))


def l1_norm(vector):
    """
    L1-norm of vector.

    :param vector: vector to get norm
    :type vector: List[float]

    :return: the l1-norm
    :rtype: float
    """
    return sum(map(abs, vector))


def l2_norm(vector):
    """
    L2-norm of vector.

    :param vector: vector to get norm
    :type vector: List[float]

    :return: the l2-norm
    :rtype: float
    """
    # better than numpy for length = 20000, nearly the same for length = 2000
    return sum(x*x for x in vector)**0.5


def lp_norm(vector, mink_power=2):
    """
    Minkowski norm of vector.

    :param vector: vector to get norm
    :type vector: List[float]

    :param mink_power: the value of exponent in Minkowski norm
    :type mink_power: float

    :return: the Minkowski norm
    :rtype: float
    """
    return float(npy.linalg.norm(vector, ord=mink_power))


def inf_norm(vector):
    """
    Infinite norm of vector.

    :param vector: vector to get norm
    :type vector: List[float]

    :return: maximum value of absolute values in vector
    :rtype: float
    """
    return max(abs(coord) for coord in vector)


def manhattan_distance(list_a, list_b):
    """
    Compute the l1 distance between list_a and list_b, i.e. the l1-norm of difference between list_a and list_b.

    :param list_a: First list
    :type list_a: List[float]

    :param list_b: Second list
    :type list_b: List[float]

    :return: the l1 distance between the two list
    :rtype: float
    """
    # faster than numpy
    return l1_norm(diff_list(list_a, list_b))


def euclidian_distance(list_a, list_b):
    """
    Compute the euclidean distance between list_a and list_b, i.e. the l2-norm of difference between list_a and list_b.

    It is the natural distance of 3D space.

    :param list_a: First list
    :type list_a: List[float]

    :param list_b: Second list
    :type list_b: List[float]

    :return: the l2 distance between the two list
    :rtype: float
    """
    # faster than numpy for length = 20000, nearly the same for length = 2000
    return l2_norm(diff_list(list_a, list_b))


def minkowski_distance(list_a, list_b, mink_power= 2):
    """
    Compute the Minkowski distance between list_a and list_b, i.e. the lp-norm of difference between list_a and list_b.

    :param list_a: First list
    :type list_a: List[float]

    :param list_b: Second list
    :type list_b: List[float]

    :param mink_power: the value of exponent in Minkowski norm
    :type mink_power: float

    :return: the Minkowski distance between the two list
    :rtype: float
    """
    # faster than sum((a - b)**p for a, b in zip(list_a, list_b))**(1/p)
    return lp_norm(npy.array(list_a)-npy.array(list_b), mink_power=mink_power)


def mean(vector):
    """
    Mean of vector.

    :param vector: vector to get mean
    :type vector: List[float]

    :return: the mean of vector
    :rtype: float
    """
    return sum(vector)/len(vector)


def variance(vector):
    """
    Variance of vector.

    :param vector: vector to get variance
    :type vector: List[float]

    :return: the variance of vector
    :rtype: float
    """
    # faster than euclidian_distance(vector, [mean(vector)] * len(vector))**2 / len(vector)
    return float(npy.var(vector))


def covariance(vector_x, vector_y):
    """
    Covariance between vector_x and vector_y.

    :param vector_x: first vector to get covariance
    :type vector_x: List[float]

    :param vector_y: second vector to get covariance
    :type vector_y: List[float]

    :return: the covariance between vector_x and vector_y
    :rtype: float
    """
    # nearly as fast as numpy
    if len(vector_x) != len(vector_y):
        raise ValueError("vector_x and vector_y must be the same length to compute covariance.")
    mean_x = mean(vector_x)
    mean_y = mean(vector_y)
    return sum((x - mean_x) * (y - mean_y) for x, y in zip(vector_x, vector_y)) / len(vector_x)


def covariance_matrix(matrix):
    """
    Compute the covariance matrix of `matrix` of dimension `N x M`.

    :return: the covariance matrix of `matrix`
    :rtype: List[List[float]], `N x N`

    :Examples:
    >>> from dessia_common.datatools.metrics import covariance_matrix
    >>> from dessia_common.models import all_cars_wi_feat
    >>> matrix = Dataset(all_cars_wi_feat).matrix
    >>> cov_matrix = covariance_matrix(list(zip(*matrix)))
    >>> for row in cov_matrix[:2]: print(row[:2])
    [70.58986267712706, -0.6737370735267286]
    [-0.6737370735267286, 0.011008722272395539]
    """
    return npy.cov(matrix, dtype=float).tolist()


def std(vector):
    """
    Standard deviation of vector.

    :param vector: vector to get standard deviation
    :type vector: List[float]

    :return: the standard deviation of vector
    :rtype: float
    """
    # faster than euclidian_distance(vector, [mean(vector)] * len(vector)) / math.sqrt(len(vector))
    return float(npy.std(vector))


def mahalanobis_distance(list_a, list_b, cov_matrix):
    """
    Compute the Mahalanobis distance between list_a and list_b.

    This method computes distances considering the scale
    and the data repartition on each dimension (covariance matrix). It is advised to use this method to compute
    distances in spaces constituted of very different dimensions in terms of scale and data repartition.

    :param list_a: First list
    :type list_a: List[float]

    :param list_b: Second list
    :type list_b: List[float]

    :param cov_matrix: the covariance matrix of data
    :type cov_matrix: List[List[float]]

    :return: the Mahalanobis distance between the two list
    :rtype: float
    """
    inv_cov_matrix = npy.linalg.pinv(cov_matrix)
    return mahalanobis(list_a, list_b, inv_cov_matrix)



""" Tools for copying objects. """

import copy
import warnings
from dessia_common.abstract import CoreDessiaObject
import dessia_common.files
from dessia_common.utils.types import is_sequence
from dessia_common.schemas.core import is_typing


def deepcopy_value(value, memo):
    """
    Returns the deep-copy of a value.

    :param value: The value to be deep-copied
    :param memo: A dictionary linking a path to an object
    :return: A deep-copy of the value
    """
    if isinstance(value, type) or is_typing(value):  # For class
        return value

    if isinstance(value, (float, int, str)):
        copied_value = value
        return copied_value

    if value is None:
        return None

    if value.__class__.__name__ in ['Point2D', 'Point3D', 'Vector2D', 'Vector3D']:
        try:
            copied_value = value.copy(deep=True, memo=memo)
        except TypeError:
            warnings.warn(f'{value.__class__.__name__}.copy() does not implement deep and memo arguments')
            copied_value = value.copy()
        return copied_value

    if isinstance(value, CoreDessiaObject):
        memo_value = search_memo(value, memo)
        if memo_value is not None:
            return memo_value
        try:
            copied_value = value.copy(deep=True, memo=memo)
        except TypeError:
            warnings.warn(f'{value.__class__.__name__}.copy() does not implement deep and memo arguments')
            copied_value = value.copy()

        memo[value] = copied_value
        return copied_value

    if isinstance(value, (dessia_common.files.BinaryFile, dessia_common.files.StringFile)):
        return value.copy()

    if hasattr(value, '__deepcopy__'):
        memo_value = search_memo(value, memo)
        if memo_value is not None:
            return memo_value

        try:
            copied_value = copy.deepcopy(value, memo=memo)
        except TypeError:
            # Memo not handled
            copied_value = copy.deepcopy(value)
        memo[value] = copied_value
        return copied_value

    if is_sequence(value):
        return deepcopy_sequence(value, memo)

    if isinstance(value, dict):
        return deepcopy_dict(value, memo)

    raise NotImplementedError(f'unhandle type for copy: {value} of type {value.__class__}')


def deepcopy_dict(dict_value, memo):
    """ Deep-copy dict. """
    memo_value = search_memo(dict_value, memo)
    if memo_value is not None:
        return memo_value

    copied_dict = {}
    for key, value in dict_value.items():
        copied_k = deepcopy_value(key, memo=memo)
        copied_v = deepcopy_value(value, memo=memo)
        copied_dict[copied_k] = copied_v
    return copied_dict


def deepcopy_sequence(seq_value, memo):
    """ Deep-copy sequence. """
    memo_value = search_memo(seq_value, memo)
    if memo_value is not None:
        return memo_value

    copied_list = []
    for value in seq_value:
        copied_value = deepcopy_value(value, memo=memo)
        copied_list.append(copied_value)
    return copied_list


def search_memo(value, memo):
    """ Search in given memo. """
    try:
        return memo[value]
    except (TypeError, KeyError):
        return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Module for errors in dessia_common. """


class ExceptionWithTraceback(Exception):
    """ Base class with a message and a traceback. """

    def __init__(self, message, traceback_=''):
        super().__init__(message)
        self.message = message
        self.traceback = traceback_

    def __str__(self):
        return f'{self.message}\nTraceback:\n{self.traceback}'


class DeepAttributeError(ExceptionWithTraceback, AttributeError):
    """ Error when trying to reach a deep attribute of an object, but couldn't get it. """


class ModelError(Exception):
    """ Error of model. """


class ConsistencyError(Exception):
    """ Error of consistency. """


class SerializationError(Exception):
    """ Error of serialization. """


class DeserializationError(Exception):
    """ Error of deserialization. """


class CopyError(Exception):
    """ Error of copy. """


class UntypedArgumentError(Exception):
    """ Error of code annotation. """



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

"""

import os
from dessia_common.models import simulation_list, system1
from dessia_common.core import DessiaObject
import dessia_common.serialization as dcs
from dessia_common.displays import draw_networkx_graph

d = simulation_list.to_dict()
simulation_list_2 = DessiaObject.dict_to_object(d)

assert simulation_list_2 == simulation_list

simulation_list_copy = simulation_list.copy()
assert simulation_list_copy == simulation_list
diff = simulation_list_copy._data_diff(simulation_list)
assert diff.is_empty()
# Let this print to test diff utils __repr__
print(diff)

simulation_list.schema()

pointer_analysis = dcs.pointers_analysis(simulation_list)
pointer_graph = dcs.pointer_graph(d)
draw_networkx_graph(pointer_graph)

system1._check_platform()
system1.schema()
system1.save_export_to_file('xlsx', 'generic_xlsx')
os.path.isfile('generic_xlsx.xlsx')

check_list = system1.check_list()

system1.save_to_file('system1')
system1_lff = DessiaObject.from_json('system1.json')
assert system1_lff == system1



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Serialization Tools. """


import warnings
import dessia_common.serialization as dcs


def serialize(value):
    """ Main function for serialization without pointers. Calls recursively itself. """
    warnings.warn("Module serialization.py have been moved outside utils. Please use it instead")
    return dcs.serialize(value)


def serialize_with_pointers(value, memo=None, path='#', id_method=True, id_memo=None):
    """ Main function for serialization with pointers. """
    warnings.warn("Module serialization.py have been moved outside utils. Please use it instead")
    return dcs.serialize_with_pointers(value=value, memo=memo, path=path, id_method=id_method, id_memo=id_memo)


def deserialize(serialized_element, sequence_annotation: str = 'List', global_dict=None,
                pointers_memo=None, path: str = '#'):
    """ Main function for deserialization, handle pointers. """
    warnings.warn("Module serialization.py have been moved outside utils. Please use it instead")
    return dcs.deserialize(serialized_element=serialized_element, sequence_annotation=sequence_annotation,
                           global_dict=global_dict, pointers_memo=pointers_memo, path=path)


def dict_to_object(dict_, class_=None, force_generic: bool = False, global_dict=None, pointers_memo=None, path='#'):
    """ Transform a dictionary to an object. """
    warnings.warn("Module serialization.py have been moved outside utils. Please use it instead")
    return dcs.dict_to_object(dict_=dict_, class_=class_, force_generic=force_generic, global_dict=global_dict,
                              pointers_memo=pointers_memo, path=path)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Utilities for workflows. """

from typing import List, Dict


class ToScriptElement:
    """ Class meant to improve to_script readability. """

    def __init__(self, declaration: str, before_declaration: str = None,
                 imports: List[str] = None, imports_as_is: List[str] = None):
        self.before_declaration = before_declaration
        self.declaration = declaration
        self.imports = imports
        self.imports_as_is = imports_as_is

    def imports_to_str(self) -> str:
        """  Import a script as str. """
        script_imports = ""
        for module, class_list in self.get_import_dict().items():
            script_imports += f"from {module} import {', '.join(class_list)}\n"

        for import_as_is in self.imports_as_is:
            import_str = f"import {import_as_is}\n"
            if import_str not in script_imports:
                script_imports += import_str

        return script_imports

    def get_import_dict(self) -> Dict[str, List[str]]:
        """ Get imported modules of script into a dict. """
        imports_dict: Dict[str, List[str]] = {}
        for import_ in self.imports:
            module = '.'.join(import_.split('.')[:-1])
            class_ = import_.split('.')[-1]
            if module not in imports_dict :
                imports_dict[module] = [class_]
            else:
                if class_ not in imports_dict[module]:
                    imports_dict[module].append(class_)

        return imports_dict


def blocks_to_script(blocks, prefix: str, imports):
    """ Set B locks to script. """
    blocks_str = ""
    for iblock, block in enumerate(blocks):
        block_script = block._to_script(prefix)
        imports.extend(block_script.imports)
        if block_script.before_declaration is not None:
            blocks_str += f"{block_script.before_declaration}\n"
        blocks_str += f'{prefix}block_{iblock} = {block_script.declaration}\n'
    blocks_str += f"{prefix}blocks = [{', '.join([prefix + 'block_' + str(i) for i in range(len(blocks))])}]\n"
    return blocks_str


def nonblock_variables_to_script(nonblock_variables, prefix, imports, imports_as_is):
    """ Set Non Block Variables to script. """
    nbvs_str = ""
    for nbv_index, nbv in enumerate(nonblock_variables):
        nbv_script = nbv._to_script()
        imports.extend(nbv_script.imports)
        imports_as_is.extend(nbv_script.imports_as_is)
        nbvs_str += f"{prefix}variable_{nbv_index} = {nbv_script.declaration}\n"
    return nbvs_str



"""
Tests for document generator
"""

from dessia_common.document_generator import DocxWriter, Paragraph, Heading, Footer, Section, Table

paragraph_1 = Paragraph(text="This is the first paragraph.")
paragraph_2 = Paragraph(text="The second paragraph is here.")
paragraph_3 = Paragraph(text="In the third paragraph, we discuss some important details.")
paragraph_4 = Paragraph(text="Paragraph number four is shorter than the others.")
paragraph_5 = Paragraph(text="The fifth paragraph is longer and more detailed than the previous ones.")
paragraph_6 = Paragraph(text="Finally, the sixth and last paragraph concludes our discussion.")
paragraphs = [paragraph_1, paragraph_2, paragraph_3, paragraph_4, paragraph_5, paragraph_6]

heading_1 = Heading(text='Heading 1', level=1)
heading_2 = Heading(text='Heading 2', level=2)
heading_3 = Heading(text='Heading 3', level=1)
heading_4 = Heading(text='Heading 4', level=2)
headings = [heading_1, heading_2, heading_3, heading_4]

footer = Footer(text="DessIA Technologies", align="left")
footer_2 = Footer(text="DessIA Technologies", align="right")
section = Section()
section.add_element(element=footer)
section.add_element(element=footer_2)

table = Table([['Library', 'Platform'], ['dessia_common', 'testing'], ['volmdlr', 'dev']])

writer = DocxWriter(filename="test_docx.docx", paragraphs=paragraphs,
                    headings=headings, section=section, tables=[table])

writer.add_headings()
writer.add_paragraphs()
writer.add_table(table_index=0)
writer.save_file()

document = writer.document

# Check the headings
assert document.paragraphs[0].text == 'Heading 1'
assert document.paragraphs[0].style.name == 'Heading 1'
assert document.paragraphs[1].text == 'Heading 2'
assert document.paragraphs[1].style.name == 'Heading 2'
assert document.paragraphs[2].text == 'Heading 3'
assert document.paragraphs[2].style.name == 'Heading 1'
assert document.paragraphs[3].text == 'Heading 4'
assert document.paragraphs[3].style.name == 'Heading 2'

# Check the paragraphs
assert document.paragraphs[6].text == 'This is the first paragraph.'
assert document.paragraphs[8].text == 'The second paragraph is here.'
assert document.paragraphs[10].text == 'In the third paragraph, we discuss some important details.'
assert document.paragraphs[12].text == 'Paragraph number four is shorter than the others.'
assert document.paragraphs[13].text == 'The fifth paragraph is longer and more detailed than the previous ones.'
assert document.paragraphs[14].text == 'Finally, the sixth and last paragraph concludes our discussion.'

# Check the table
table = document.tables[0]
assert table.cell(0, 0).text == 'Library'
assert table.cell(0, 1).text == 'Platform'
assert table.cell(1, 0).text == 'dessia_common'
assert table.cell(1, 1).text == 'testing'
assert table.cell(2, 0).text == 'volmdlr'
assert table.cell(2, 1).text == 'dev'



import pkg_resources
from dessia_common.models.workflows.workflow_from_file_input import workflow_
from dessia_common.files import StringFile


stream = StringFile.from_stream(pkg_resources.resource_stream('dessia_common', 'models/data/seed_file.csv'))

string_file = StringFile("test_file_input")
string_file.write(stream.read())
inputs = {0: string_file}
workflow_run = workflow_.run(input_values=inputs)
assert workflow_run.output_value.standalone_subobject.intarg == 7
workflow_run._check_platform()



""" Test Workflow to check good behavior with file inputs. """

from dessia_common.workflow.core import Workflow
from dessia_common.workflow.blocks import ClassMethod
from dessia_common.utils.types import ClassMethodType
from dessia_common.forms import StandaloneObject

method_type = ClassMethodType(class_=StandaloneObject, name="generate_from_text")
class_method = ClassMethod(method_type=method_type, name="Class Method")

blocks = [class_method]
pipes = []

workflow_ = Workflow(blocks=blocks, pipes=pipes, output=class_method.outputs[0])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import dessia_common.displays as dcd
import dessia_common.graph as dcg
import networkx as nx

G = nx.gn_graph(80)

dcd.draw_networkx_graph(G)

reg = dcg.extract_region(G, [1, 2])

assert 1 in reg

order = dcg.explore_tree_from_leaves(G)
graph2 = dcg.cut_tree_final_branches(reg)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Exports for dessia_common. """
from typing import List, Dict, Any, Sequence, Optional


from openpyxl.styles.borders import Border, Side
from openpyxl.styles import PatternFill, Font
from openpyxl import Workbook
import openpyxl.utils
from dessia_common.utils.types import is_sequence
from dessia_common.breakdown import breakdown


def is_hashable(value):
    """ Determine whether `value` can be hashed. """
    try:
        hash(value)
    except TypeError:
        return False
    return True


def is_number(value):
    """ Determine if the value is a int or a float. """
    return isinstance(value, (int, float))


def is_builtins_list(list_):
    """ Determine if a list is only composed of builtins. """
    for element in list_:
        if not (is_number(element) or isinstance(element, str)):
            return False
    return True


class ExportFormat:
    """ Define which method of an object should be called for each Export. """

    def __init__(self, selector: Optional[str], extension: str, method_name: str, text: bool,
                 export_name: str = "", args: Dict[str, Any] = None):
        self.selector = selector
        self.extension = extension
        self.method_name = method_name
        self.text = text
        self.export_name = export_name
        if args is None:
            args = {}
        self.args = args

    def to_dict(self):
        """ Serialization. """
        return {"selector": self.selector, "extension": self.extension, "method_name": self.method_name,
                "text": self.text, "export_name": self.export_name, "args": self.args}


class XLSXWriter:
    """ Base class to write a DessiaObject in an excel file. """

    max_column_width = 40
    color_dessIA1 = "263238"
    color_dessIA2 = "537CB0"
    grey1 = "f1f1f1"
    white_font = Font(color="FFFFFF")

    thin_border = Border(left=Side(style='thin'), right=Side(style='thin'),
                         top=Side(style='thin'), bottom=Side(style='thin'))

    def __init__(self, object_):
        self.pattern_color1 = PatternFill(
            fill_type="solid",
            start_color=self.color_dessIA1,
            end_color=self.color_dessIA1)

        self.pattern_color2 = PatternFill(
            fill_type="solid",
            start_color=self.color_dessIA2,
            end_color=self.color_dessIA2)

        self.workbook = Workbook()
        self.main_sheet = self.workbook.active
        self.object = object_

        self.paths = breakdown(object_)

        self.classes_to_sheets = {}
        self.object_to_sheet_row = {}
        for class_name, obj_paths in self.paths.items():
            sheet = self.workbook.create_sheet(class_name)
            self.classes_to_sheets[class_name] = sheet

            for i, (obj, path) in enumerate(obj_paths.items()):
                self.object_to_sheet_row[obj] = (sheet, i + 2, path)

        self.write()

    def write_class_header_to_row(self, obj_of_class, sheet, row_number):
        """ Write to a sheet the class header: finds columns names from a class. """
        cell = sheet.cell(row=row_number, column=1, value='Path')
        cell.fill = self.pattern_color2
        cell.border = self.thin_border
        cell.font = self.white_font

        cell = sheet.cell(row=row_number, column=2, value='name')
        cell.fill = self.pattern_color2
        cell.border = self.thin_border
        cell.font = self.white_font
        i = 3

        for (k, _) in sorted(obj_of_class.__dict__.items()):
            if (not k.startswith('_')) and k != 'name':
                cell = sheet.cell(row=row_number, column=i, value=str(k))

                cell.border = self.thin_border
                cell.fill = self.pattern_color2
                cell.font = self.white_font
                i += 1

    def write_value_to_cell(self, value, sheet, row_number, column_number):
        """ Write a given value to a cell. Insert it as a link if it is an object. """
        cell_link = None
        if isinstance(value, dict):
            str_v = f'Dict of {len(value)} items'
        elif isinstance(value, list):
            if is_builtins_list(value):
                str_v = str(value)
            else:
                str_v = f'List of {len(value)} items'

        elif isinstance(value, set):
            str_v = f'Set of {len(value)} items'
        elif isinstance(value, float):
            str_v = round(value, 6)
        elif is_hashable(value) and value in self.object_to_sheet_row:
            ref_sheet, ref_row_number, ref_path = self.object_to_sheet_row[value]
            str_v = ref_path
            cell_link = f'#{ref_sheet.title}!A{ref_row_number}'
        else:
            str_v = str(value)

        cell = sheet.cell(row=row_number, column=column_number, value=str_v)
        if cell_link:
            cell.hyperlink = cell_link

        cell.border = self.thin_border

    def write_object_to_row(self, obj, sheet, row_number, path=''):
        """ Write on object to a row. Loops on its attributes to write its value in each cell. """
        cell = sheet.cell(row=row_number, column=1, value=path)
        cell.border = self.thin_border
        if hasattr(obj, 'name'):
            cell = sheet.cell(row=row_number, column=2, value=obj.name)
        else:
            cell = sheet.cell(row=row_number, column=2, value='No name in model')

        cell.border = self.thin_border
        i = 3
        for (k, value) in sorted(obj.__dict__.items()):
            if (not k.startswith('_')) and k != 'name':
                self.write_value_to_cell(value, sheet, row_number, i)

                i += 1

                # column_width = min((len(k) + 1.5), max_column_width)
                # column_name = openpyxl.utils.cell.get_column_letter(i)
                # sheet.column_dimensions[column_name].width = column_width

    def write_object_id(self, sheet):
        """ Write object id to a given sheet. """
        sheet.title = f'Object {self.object.__class__.__name__}'

        sheet['A1'] = 'Module'
        sheet['B1'] = 'Class'
        sheet['C1'] = 'name'

        sheet['A1'].border = self.thin_border
        sheet['B1'].border = self.thin_border
        sheet['C1'].border = self.thin_border
        sheet['A1'].fill = self.pattern_color1
        sheet['B1'].fill = self.pattern_color1
        sheet['C1'].fill = self.pattern_color1
        sheet['A1'].font = self.white_font
        sheet['B1'].font = self.white_font
        sheet['C1'].font = self.white_font

        sheet['A2'] = self.object.__module__
        sheet['B2'] = self.object.__class__.__name__
        sheet['C2'] = self.object.name

        sheet['A2'].border = self.thin_border
        sheet['B2'].border = self.thin_border
        sheet['C2'].border = self.thin_border
        sheet['A2'].fill = self.pattern_color1
        sheet['B2'].fill = self.pattern_color1
        sheet['C2'].fill = self.pattern_color1
        sheet['A2'].font = self.white_font
        sheet['B2'].font = self.white_font
        sheet['C2'].font = self.white_font

        sheet['A3'] = 'Attribute'
        sheet['A4'] = 'Value'
        sheet['A3'].border = self.thin_border
        sheet['A4'].border = self.thin_border

    def write(self):
        """ Generate the whole file. """
        # name_column_width = 0
        self.write_object_id(self.main_sheet)
        self.write_class_header_to_row(self.object, self.main_sheet, 3)
        self.write_object_to_row(self.object, self.main_sheet, 4)
        self.autosize_sheet_columns(self.main_sheet, 5, 30)

        for class_name, obj_paths in self.paths.items():
            sheet = self.classes_to_sheets[class_name]

            for obj, path in obj_paths.items():
                _, row_number, path = self.object_to_sheet_row[obj]
                self.write_object_to_row(obj, sheet, row_number, path)
            self.write_class_header_to_row(obj, sheet, 1)

            sheet.auto_filter.ref = f"A1:{openpyxl.utils.cell.get_column_letter(sheet.max_column)}{len(obj_paths) + 1}"
            self.autosize_sheet_columns(sheet, 5, 30)

    def save_to_file(self, filepath: str):
        """ Save to a filepath (open) and write. """
        if not filepath.endswith('.xlsx'):
            filepath += '.xlsx'
            print(f"Changing name to {filepath}")

        with open(filepath, 'rb') as file:
            self.save_to_stream(file)

    def save_to_stream(self, stream):
        """ Saves the file to a binary stream. """
        self.workbook.save(stream)

    @staticmethod
    def autosize_sheet_columns(sheet, min_width=5, max_width=30):
        """ Auto-size the sheet columns by analyzing the content. Min and max width must be specified. """
        for col in sheet.columns:
            width = min_width
            column = col[1].column_letter  # Get the column name
            # Since Openpyxl 2.6, the column name is  ".column_letter" as .column became the column number (1-based)
            for cell in col:
                try:  # Necessary to avoid error on empty cells
                    if len(str(cell.value)) > width:
                        width = len(str(cell.value))
                except AttributeError:
                    pass
            if width > 0:
                adjusted_width = min((width + 0.5), max_width)
                sheet.column_dimensions[column].width = adjusted_width


class MarkdownWriter:
    """ Base class to write markdowns. """

    def __init__(self, print_limit: int = 25, table_limit: Optional[int] = 12):
        self.print_limit = print_limit
        self.table_limit = table_limit

    @staticmethod
    def object_titles():
        """ Return a list of strings representing the titles for the object matrix. """
        return ['Attribute', 'Type', 'Value']

    @staticmethod
    def _sequence_to_str(value: Sequence):
        if len(value) == 0:
            return f"empty {type(value).__name__}"

        printed_string = f"{len(value)} "

        all_class_names = list(set(subvalue.__class__.__name__ for subvalue in value))
        if len(all_class_names) == 1:
            printed_string += ''.join([f"{all_class_names[0]}",
                                       f"{'s' if len(value) > 1 else ''}"])
        else:
            str_all_class = str(all_class_names).translate(str(all_class_names).maketrans('', '', "{}'"))
            printed_string += f"{str_all_class}"

        return printed_string

    def _dict_to_str(self, value: Dict):
        return self._sequence_to_str(list(value.values()))

    @staticmethod
    def _object_to_str(value) -> str:
        if hasattr(value, 'name') and value.name:
            return value.name
        return 'unnamed'

    def _value_to_str(self, value: Any) -> str:
        if isinstance(value, (float, int, bool, complex)):
            return str(round(value, 6))
        if isinstance(value, str):
            return value if value != '' else 'no value'
        if is_sequence(value):
            return self._sequence_to_str(value)
        if isinstance(value, Dict):
            return self._dict_to_str(value)
        if value is None:
            return ' - '
        return self._object_to_str(value)

    def _string_in_table(self, string: str = ''):
        return string[:self.print_limit] + ('...' if len(string) > self.print_limit else '')

    def object_matrix(self, object_):
        """ Return a matrix representing the object passed as argument. """
        matrix = []
        for attr, value in object_.__dict__.items():
            matrix.append([attr,
                           value.__class__.__name__,
                           self._value_to_str(value)])
        return matrix

    @staticmethod
    def _head_table(col_names: List[str]) -> str:
        cap_names = map(lambda x: x.capitalize(), col_names)
        return ("| " + " | ".join(cap_names) + " |\n" +
                "| ------ " * len(col_names) + "|\n")

    def _table_line(self, row: List[Any]) -> str:
        line = "|"
        for value in row:
            line += f" {self._string_in_table(self._value_to_str(value))} |"
        return line + "\n"

    def _table_rows_from_content(self, content: List[List[Any]]) -> str:
        string = ''
        for row in content:
            string += self._table_line(row)
        return string

    def _content_table(self, content: List[List[Any]]) -> str:
        if self.table_limit is None:
            return self._table_rows_from_content(content)

        table = ''
        half_table = int(self.table_limit / 2)

        table += self._table_rows_from_content(content[:half_table])

        if self.table_limit > 1:
            if len(content) > self.table_limit:
                table += f"| + {len(content) - self.table_limit} unprinted elements | |\n"

            if len(content) > half_table:
                table += self._table_rows_from_content(content[-half_table:])

        return table

    @staticmethod
    def print_name(object_) -> str:
        """Print name of object_."""
        return object_.name if object_.name != '' else 'with no name'

    @staticmethod
    def print_class(object_) -> str:
        """Print name of class name of object_."""
        return object_.__class__.__name__

    def matrix_table(self, matrix: List[List[float]], col_names: List[str]) -> str:
        """Print col_names of matrix as a table."""
        return ''.join([self._head_table(col_names),
                        self._content_table(matrix)])

    def object_table(self, object_) -> str:
        """Print object_'s attributes in table."""
        return self.matrix_table(self.object_matrix(object_),
                                 self.object_titles())

    def element_details(self, elements: List[Any]) -> str:
        """Print sequence of elements."""
        return self._sequence_to_str(elements)

    @staticmethod
    def write_to_file(filename: str, content: str) -> None:
        """ Writes the given content to the specified file. """
        with open(filename, 'w', encoding='utf-8') as file:
            file.write(content)

    @staticmethod
    def table_of_contents(headings: List[str]) -> str:
        """ Generates a table of contents based on the given list of headings. """
        table_of_contents = '## Table of Contents\n\n'
        for heading in headings:
            table_of_contents += f'- [{heading}](#{heading.lower().replace(" ", "-")})\n'
        return table_of_contents

    @staticmethod
    def header(title: str, level: int = 1) -> str:
        """ Generates a markdown header with the specified title and level. """
        header_level = "#" * level
        return f"{header_level} {title}\n\n"



from dessia_common.workflow.core import Workflow, Pipe
from dessia_common.workflow.blocks import Unpacker, Sequence, InstantiateModel, ModelMethod
from dessia_common.typings import MethodType

from dessia_common.forms import Generator

generate = InstantiateModel(model_class=Generator, name="Test1")
method_type = MethodType(class_=Generator, name="generate")
method = ModelMethod(method_type=method_type)

unpack1 = Unpacker(indices=[0, 2])
unpack2 = Unpacker(indices=[1, 3])

sequence = Sequence(number_arguments=4)

blocks = [generate, method, unpack1, unpack2, sequence]
pipes = [Pipe(input_variable=generate.outputs[0], output_variable=method.inputs[0]),
         Pipe(input_variable=method.outputs[0], output_variable=unpack1.inputs[0]),
         Pipe(input_variable=method.outputs[0], output_variable=unpack2.inputs[0]),
         Pipe(input_variable=unpack1.outputs[0], output_variable=sequence.inputs[0]),
         Pipe(input_variable=unpack1.outputs[1], output_variable=sequence.inputs[2]),
         Pipe(input_variable=unpack2.outputs[0], output_variable=sequence.inputs[1]),
         Pipe(input_variable=unpack2.outputs[1], output_variable=sequence.inputs[3])]

workflow_ = Workflow(blocks=blocks, pipes=pipes, output=sequence.outputs[0])

assert len(workflow_.blocks) == len(workflow_.runtime_blocks) and set(workflow_.runtime_blocks) == set(workflow_.blocks)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Module to compute differences between objects. """

import math
from typing import List

import numpy as npy
from dessia_common import FLOAT_TOLERANCE
import dessia_common.core as dc
from dessia_common.utils.types import isinstance_base_types, is_sequence
from dessia_common.utils.helpers import full_classname
from dessia_common.files import BinaryFile, StringFile


class DifferentValues:
    """ Contains info on diff different values. """

    def __init__(self, path, value1, value2):
        self.path = path
        self.value1 = value1
        self.value2 = value2

    def __repr__(self):
        return f'{self.path}: values differ {self.value1} / {self.value2}'


class MissingAttribute:
    """ Contains info on diff missing attribute. """

    def __init__(self, path: str, missing_in_first_object: bool):
        self.path = path
        self.missing_in_first_object = missing_in_first_object

    def __repr__(self):
        if self.missing_in_first_object:
            return f'{self.path}: missing in first object'

        return f'{self.path}: missing in second object'


class DifferentType:
    """ Contains info on diff different types. """

    def __init__(self, path: str, value1, value2):
        self.path = path
        self.value1 = value1
        self.value2 = value2

    def __repr__(self):
        return f'{self.path}: types differ {self.value1} / {self.value2}'


class Diff:
    """ Contains info on a performed diff analysis between two values. """

    def __init__(self, different_values: List[DifferentValues], missing_attributes: List[MissingAttribute],
                 invalid_types: List[DifferentType]):
        self.different_values = different_values
        self.missing_attributes = missing_attributes
        self.invalid_types = invalid_types

    def is_empty(self):
        """ Return True if diff is empty, else False. """
        return (not self.different_values) and (not self.missing_attributes) and (not self.invalid_types)

    def __repr__(self):
        if self.is_empty():
            return 'Objects are equal, diff is empty'

        object_print = ''
        if self.different_values:
            object_print += 'Different values:\t'
            for diff_value in self.different_values:
                object_print += f'\t{diff_value}\n'

        if self.missing_attributes:
            object_print += 'Missing attributes:\t'
            for diff_value in self.missing_attributes:
                object_print += f'\t{diff_value}\n'

        if self.invalid_types:
            object_print += 'Invalid types:\n'
            for diff_value in self.invalid_types:
                object_print += f'\t{diff_value}\n'
        return object_print

    def __add__(self, other_diff):
        return Diff(self.different_values + other_diff.different_values,
                    self.missing_attributes + other_diff.missing_attributes,
                    self.invalid_types + other_diff.invalid_types)


def diff(value1, value2, path='#'):
    """
    Main function to get the diff of two objects.

    :return: a tuple of a list of diff between the objects, missing keys in other object and invalid types
    """
    diff_values = []
    missing_keys_in_other_object = []
    invalid_types = []

    if is_sequence(value1) and is_sequence(value2):
        return sequence_diff(value1, value2, path=path)

    if not isinstance(value1, type(value2)):
        invalid_types.append(DifferentType(path, value1, value2))
        return Diff(diff_values, missing_keys_in_other_object, invalid_types)

    if isinstance_base_types(value1):
        if isinstance(value1, float) and math.isclose(value1, value2, abs_tol=FLOAT_TOLERANCE):
            return Diff(diff_values, missing_keys_in_other_object, invalid_types)
        if value1 != value2:
            # diff_values.append((path, value1, value2))
            diff_values.append(DifferentValues(path, value1, value2))
        return Diff(diff_values, missing_keys_in_other_object, invalid_types)
    if isinstance(value1, dict):
        return dict_diff(value1, value2, path=path)

    # Should be object
    if hasattr(value1, '_data_eq'):
        # DessiaObject
        if value1._data_eq(value2):
            return Diff([], [], [])

        # Use same code snippet as in data_eq
        eq_dict = value1._serializable_dict()
        if 'name' in eq_dict:
            del eq_dict['name']

        other_eq_dict = value2._serializable_dict()
        return dict_diff(eq_dict, other_eq_dict, path=path)

    if value1 == value2:
        return Diff([], [], [])

    raise NotImplementedError(f'Undefined type in diff: {type(value1)}')


def dict_diff(dict1, dict2, path='#'):
    """ Returns diff between two dictionaries. """
    diff_object = Diff([], [], [])

    dict_ = dict1
    other_dict = dict2
    first_object = False
    if len(dict2) > len(dict1):
        dict_ = dict2
        other_dict = dict1
        first_object = True
    for key, value in dict_.items():
        path_key = f'{path}/{key}'
        if key not in other_dict:
            diff_object.missing_attributes.append(MissingAttribute(path=path_key,
                                                                   missing_in_first_object=first_object))
        else:
            diff_key = diff(value, other_dict[key], path=path_key)
            diff_object += diff_key
    return diff_object


def sequence_diff(seq1, seq2, path='#'):
    """ Returns diff between two sequences. """
    seq_diff = Diff([], [], [])

    if len(seq1) != len(seq2):
        # diff_values.append((path, seq1, seq2))
        seq_diff.different_values.append(DifferentValues(path, seq1, seq2))
    else:
        for i, (v1, v2) in enumerate(zip(seq1, seq2)):
            path_value = f'{path}/{i}'
            diff_value = diff(v1, v2, path=path_value)
            seq_diff += diff_value
    return seq_diff


def data_eq(value1, value2):
    """ Returns if two values are equal on data equality. """
    if is_sequence(value1) and is_sequence(value2):
        return sequence_data_eq(value1, value2)

    if isinstance(value1, npy.int64) or isinstance(value2, npy.int64):
        return value1 == value2

    if isinstance(value1, npy.float64) or isinstance(value2, npy.float64):
        return math.isclose(value1, value2, abs_tol=FLOAT_TOLERANCE)

    if not isinstance(value2, type(value1))\
            and not isinstance(value1, type(value2)):
        return False

    if isinstance_base_types(value1):
        if isinstance(value1, float):
            return math.isclose(value1, value2, abs_tol=FLOAT_TOLERANCE)

        return value1 == value2

    if isinstance(value1, dict):
        return dict_data_eq(value1, value2)

    if isinstance(value1, (BinaryFile, StringFile)):
        return value1 == value2

    if isinstance(value1, type):
        return full_classname(value1) == full_classname(value2)

    # Else: its an object
    if full_classname(value1) != full_classname(value2):
        return False

    # Test if _data_eq is customized
    if hasattr(value1, '_data_eq'):
        custom_method = value1._data_eq.__code__ is not dc.DessiaObject._data_eq.__code__
        if custom_method:
            return value1._data_eq(value2)

    # Not custom, use generic implementation
    eq_dict = value1._data_eq_dict()
    if 'name' in eq_dict:
        del eq_dict['name']

    other_eq_dict = value2._data_eq_dict()

    return dict_data_eq(eq_dict, other_eq_dict)


def dict_data_eq(dict1, dict2):
    """ Returns True if two dictionaries are equal on data equality, False otherwise. """
    for key, value in dict1.items():
        if key not in dict2:
            return False
        if not data_eq(value, dict2[key]):
            return False
    return True


def sequence_data_eq(seq1, seq2):
    """ Returns if two sequences are equal on data equality. """
    if len(seq1) != len(seq2):
        return False

    for v1, v2 in zip(seq1, seq2):
        if not data_eq(v1, v2):
            return False

    return True


def choose_hash(object_):
    """ Base function to return hash. """
    if is_sequence(object_):
        return sequence_hash(object_)
    if isinstance(object_, dict):
        return dict_hash(object_)
    if isinstance(object_, str):
        return sum(ord(e) for e in object_)
    return hash(object_)


def sequence_hash(sequence):
    """
    Return hash of a sequence value.

    Only checks for first and last elements hashes if defined for performance purpose.
    It also looks that previous sequence hash method was less efficient as the sum of all hashes in sequence
    returned less unique values for normally different sequences.
    """
    if not sequence:
        return 0

    # Recursively compute hash of first and last element for performance purpose
    hash_ = len(sequence)*choose_hash(sequence[0])
    if len(sequence) > 1:
        hash_ += 5381*choose_hash(sequence[-1])
    return hash_


def dict_hash(dict_):
    """
    Returns hash of a dict value.

    If keys can be ordered, only checks for first and last elements hashes if defined for performance purpose.
    """
    if not dict_:
        return 0

    hash_ = 0
    try:
        # Try and sort keys in order to get first and last elements.
        sorted_keys = sorted(dict_.keys())
    except TypeError:
        # Old less efficient hash for non order-able keys.
        for key, value in dict_.items():
            hash_ += hash(key) + choose_hash(value)
        return hash_

    # Recursively compute hash of first and last element for performance purpose
    first_key = sorted_keys[0]
    hash_ += len(dict_) * choose_hash(dict_[first_key])
    if len(dict_) > 1:
        last_key = sorted_keys[-1]
        hash_ += 313 * choose_hash(dict_[last_key])
    return hash_



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jul 21 14:43:53 2021

@author: steven
"""

from dessia_common.utils.diff import dict_diff


ds1 = {'key': 'val1'}
ds2 = {'key': 'val2'}

dv1 = dict_diff(ds1, ds2).different_values[0]
assert dv1.path == '#/key'
assert dv1.value1 == 'val1'
assert dv1.value2 == 'val2'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Apr 13 17:35:52 2022

@author: steven
"""

import dessia_common.breakdown as bd
from dessia_common.models import simulation1

assert bd.attrmethod_getter(simulation1, 'system.output_power(1)') == 0.8

bda = bd.breakdown_analysis(simulation1)

assert bda['total_size'] > 20  # Actual value is 48, this test just to ensure that value is not too low



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Module to define Blocks for workflows. """

import inspect
import warnings
from zipfile import ZipFile
from typing import List, Type, Any, Dict, Tuple, get_type_hints, TypeVar, Optional
import itertools
from dessia_common.core import DessiaFilter, FiltersList, type_from_annotation, DessiaObject
from dessia_common.schemas.core import split_argspecs, parse_docstring, EMPTY_PARSED_ATTRIBUTE
from dessia_common.displays import DisplaySetting, DisplayObject
from dessia_common.errors import UntypedArgumentError
from dessia_common.typings import (JsonSerializable, MethodType, ClassMethodType, AttributeType, ViewType, CadViewType,
                                   PlotDataType, MarkdownType)
from dessia_common.files import StringFile, BinaryFile, generate_archive
from dessia_common.utils.helpers import concatenate, full_classname
from dessia_common.breakdown import attrmethod_getter, get_in_object_from_path
from dessia_common.exports import ExportFormat
from dessia_common.workflow.core import Block, Variable, TypedVariable, TypedVariableWithDefaultValue, Workflow
from dessia_common.workflow.utils import ToScriptElement

T = TypeVar("T")


def set_inputs_from_function(method, inputs=None):
    """ Inspect given method argspecs and sets block inputs from it. """
    if inputs is None:
        inputs = []
    args_specs = inspect.getfullargspec(method)
    nargs, ndefault_args = split_argspecs(args_specs)

    for iarg, argument in enumerate(args_specs.args):
        if argument not in ['self', 'cls', 'progress_callback']:
            try:
                annotations = get_type_hints(method)
                type_ = type_from_annotation(annotations[argument], module=method.__module__)
            except KeyError as error:
                raise UntypedArgumentError(f"Argument {argument} of method/function {method.__name__} has no typing")\
                    from error
            if iarg > nargs - ndefault_args:
                default = args_specs.defaults[ndefault_args - nargs + iarg - 1]
                input_ = TypedVariableWithDefaultValue(type_=type_, default_value=default, name=argument)
                inputs.append(input_)
            else:
                inputs.append(TypedVariable(type_=type_, name=argument))
    return inputs


def output_from_function(function, name: str = "result output"):
    """ Inspect given function argspecs and compute block output from it. """
    annotations = get_type_hints(function)
    if 'return' in annotations:
        type_ = type_from_annotation(annotations['return'], function.__module__)
        return TypedVariable(type_=type_, name=name)
    return Variable(name=name)


def set_block_variable_names_from_dict(func):
    """ Inspect function arguments to compute black variable names. """
    def func_wrapper(cls, dict_):
        obj = func(cls, dict_)
        if 'input_names' in dict_:
            for input_name, input_ in zip(dict_['input_names'], obj.inputs):
                input_.name = input_name
        if 'output_names' in dict_:
            output_items = zip(dict_['output_names'], obj.outputs)
            for output_name, output_ in output_items:
                output_.name = output_name
        return obj
    return func_wrapper


class BlockError(Exception):
    """ Specific BlockError Exception. """


class InstantiateModel(Block):
    """
    Instantiate given class during workflow execution.

    :param model_class: Class to instantiate.
    :param name: Name of the block.
    :param position: Position of the block in canvas.
    """

    def __init__(self, model_class: Type, name: str = '', position: Tuple[float, float] = None):
        self.model_class = model_class
        inputs = []
        inputs = set_inputs_from_function(self.model_class.__init__, inputs)
        outputs = [TypedVariable(type_=self.model_class, name='Instanciated object')]
        Block.__init__(self, inputs, outputs, name=name, position=position)

    def equivalent_hash(self):
        """ Custom hash function. Related to 'equivalent' method. """
        return len(self.model_class.__name__)

    def equivalent(self, other):
        """ Return whether the block is equivalent to the other given or not. """
        classname = self.model_class.__class__.__name__
        other_classname = other.model_class.__class__.__name__
        return Block.equivalent(self, other) and classname == other_classname

    def evaluate(self, values, **kwargs):
        """ Instantiate a model of given class with arguments that are in values. """
        arguments = {var.name: values[var] for var in self.inputs}
        return [self.model_class(**arguments)]

    def package_mix(self):
        """ Add block contribution to workflow's package_mix. """
        return {self.model_class.__module__.split('.')[0]: 1}

    def _docstring(self):
        """ Parse given class' docstring. """
        docstring = self.model_class.__doc__
        annotations = get_type_hints(self.model_class.__init__)
        parsed_docstring = parse_docstring(docstring=docstring, annotations=annotations)
        parsed_attributes = parsed_docstring["attributes"]
        block_docstring = {i: parsed_attributes[i.name] if i.name in parsed_attributes
                           else EMPTY_PARSED_ATTRIBUTE for i in self.inputs}
        return block_docstring

    def _to_script(self, _) -> ToScriptElement:
        """ Write block config into a chunk of script. """
        script = f"InstantiateModel(model_class=" \
                 f"{self.model_class.__name__}, {self.base_script()})"
        imports = [full_classname(object_=self.model_class, compute_for='class'), self.full_classname]
        return ToScriptElement(declaration=script, imports=imports)


class ClassMethod(Block):
    """
    Run given class method during workflow execution. Handle static method as well.

    :param method_type: Denotes the class and method to run.
    :param name: Name of the block.
    :param position: Position of the block in canvas.
    """

    _non_serializable_attributes = ["method"]

    def __init__(self, method_type: ClassMethodType[Type], name: str = '', position: Tuple[float, float] = None):
        self.method_type = method_type
        inputs = []

        self.method = method_type.get_method()
        inputs = set_inputs_from_function(self.method, inputs)

        self.argument_names = [i.name for i in inputs]
        output_name = f"method result of {method_type.name}"
        output = output_from_function(function=self.method, name=output_name)
        Block.__init__(self, inputs, [output], name=name, position=position)

    def equivalent_hash(self):
        """ Custom hash function. Related to 'equivalent' method. """
        classname = self.method_type.class_.__name__
        return len(classname) + 7 * len(self.method_type.name)

    def equivalent(self, other: 'ClassMethod'):
        """ Return whether the block is equivalent to the other given or not. """
        classname = self.method_type.class_.__name__
        other_classname = other.method_type.class_.__name__
        same_class = classname == other_classname
        same_method = self.method_type.name == other.method_type.name
        return Block.equivalent(self, other) and same_class and same_method

    def evaluate(self, values, **kwargs):
        """ Run given classmethod with arguments that are in values. """
        arguments = {arg_name: values[var] for arg_name, var in zip(self.argument_names, self.inputs) if var in values}
        return [self.method(**arguments)]

    def _docstring(self):
        """ Parse given method's docstring. """
        docstring = self.method.__doc__
        annotations = get_type_hints(self.method)
        parsed_docstring = parse_docstring(docstring=docstring, annotations=annotations)
        parsed_attributes = parsed_docstring["attributes"]
        block_docstring = {i: parsed_attributes[i.name] if i.name in parsed_attributes
                           else EMPTY_PARSED_ATTRIBUTE for i in self.inputs}
        return block_docstring

    def _to_script(self, _) -> ToScriptElement:
        """ Write block config into a chunk of script. """
        script = f"ClassMethod(method_type=ClassMethodType(" \
                 f"{self.method_type.class_.__name__}, '{self.method_type.name}')" \
                 f", {self.base_script()})"

        imports = [full_classname(object_=self.method_type, compute_for='instance'),
                   full_classname(object_=self.method_type.class_, compute_for='class'),
                   self.full_classname]
        return ToScriptElement(declaration=script, imports=imports)

    @classmethod
    def dict_to_object(cls, dict_: JsonSerializable, force_generic: bool = False, global_dict=None,
                       pointers_memo: Dict[str, Any] = None, path: str = '#'):
        """ Backward compatibility for old versions of blocks. """
        # Backward compatibility dessia_common < 0.14.0
        if "object_class" not in dict_["method_type"]:
            dict_["method_type"]["object_class"] = "dessia_common.typings.ClassMethodType"
        return super().dict_to_object(dict_=dict_, force_generic=True, global_dict=global_dict,
                                      pointers_memo=pointers_memo, path=path)


class ModelMethod(Block):
    """
    Run given method during workflow execution.

    :param method_type: Denotes the class and method to run.
    :param name: Name of the block.
    :param position: Position of the block in canvas.
    """

    _non_serializable_attributes = ["method"]

    def __init__(self, method_type: MethodType[Type], name: str = '', position: Tuple[float, float] = None):
        self.method_type = method_type
        inputs = [TypedVariable(type_=method_type.class_, name='model at input')]
        self.method = method_type.get_method()
        inputs = set_inputs_from_function(self.method, inputs)

        # Storing argument names
        self.argument_names = [i.name for i in inputs[1:]]

        return_output_name = f"method result of {method_type.name}"
        return_output = output_from_function(function=self.method, name=return_output_name)

        model_output_name = f"model at output {method_type.name}"
        model_output = TypedVariable(type_=method_type.class_, name=model_output_name)
        outputs = [return_output, model_output]
        if name == "":
            name = f"Model method: {method_type.name}"
        Block.__init__(self, inputs, outputs, name=name, position=position)

    def equivalent_hash(self):
        """ Custom hash function. Related to 'equivalent' method. """
        classname = self.method_type.class_.__name__
        return len(classname) + 7 * len(self.method_type.name)

    def equivalent(self, other: 'ModelMethod'):
        """ Return whether the block is equivalent to the other given or not. """
        classname = self.method_type.class_.__name__
        other_classname = other.method_type.class_.__name__
        same_model = classname == other_classname
        same_method = self.method_type.name == other.method_type.name
        return Block.equivalent(self, other) and same_model and same_method

    def evaluate(self, values, progress_callback=lambda x: None, **kwargs):
        """ Run given method with arguments that are in values. """
        arguments = {n: values[v] for n, v in zip(self.argument_names, self.inputs[1:]) if v in values}
        method = getattr(values[self.inputs[0]], self.method_type.name)
        try:
            # Trying to inject progress callback to method
            result = method(progress_callback=progress_callback, **arguments)
        except TypeError:
            result = method(**arguments)

        return [result, values[self.inputs[0]]]

    def package_mix(self):
        """ Add block contribution to workflow's package_mix. """
        return {self.method_type.class_.__module__.split('.')[0]: 1}

    def _docstring(self):
        """ Parse given method's docstring. """
        docstring = self.method.__doc__
        annotations = get_type_hints(self.method)
        parsed_docstring = parse_docstring(docstring=docstring, annotations=annotations)
        parsed_attributes = parsed_docstring["attributes"]
        block_docstring = {i: parsed_attributes[i.name] if i.name in parsed_attributes
                           else EMPTY_PARSED_ATTRIBUTE for i in self.inputs}
        return block_docstring

    def _to_script(self, _) -> ToScriptElement:
        """ Write block config into a chunk of script. """
        script = f"ModelMethod(method_type=MethodType(" \
                 f"{self.method_type.class_.__name__}, '{self.method_type.name}')" \
                 f", {self.base_script()})"

        imports = [full_classname(object_=self.method_type, compute_for='instance'),
                   full_classname(object_=self.method_type.class_, compute_for='class'),
                   self.full_classname]
        return ToScriptElement(declaration=script, imports=imports)

    @classmethod
    def dict_to_object(cls, dict_: JsonSerializable, force_generic: bool = False, global_dict=None,
                       pointers_memo: Dict[str, Any] = None, path: str = '#'):
        """ Backward compatibility for old versions of blocks. """
        # Backward compatibility dessia_common < 0.14.0
        if "object_class" not in dict_["method_type"]:
            dict_["method_type"]["object_class"] = "dessia_common.typings.MethodType"
        return super().dict_to_object(dict_=dict_, force_generic=True, global_dict=global_dict,
                                      pointers_memo=pointers_memo, path=path)


class Sequence(Block):
    """
    Concatenate n inputs into a sequence.

    :param number_arguments: Number of inputs to be concatenated.
    :param name: Block name.
    :param position: Position in canvas.
    """

    def __init__(self, number_arguments: int, name: str = '', position: Tuple[float, float] = None):
        self.number_arguments = number_arguments
        inputs = [Variable(name=f"Sequence element {i}") for i in range(self.number_arguments)]
        outputs = [TypedVariable(type_=List[T], name='sequence')]
        Block.__init__(self, inputs, outputs, name=name, position=position)

    def equivalent_hash(self):
        """ Custom hash function. Related to 'equivalent' method. """
        return self.number_arguments

    def equivalent(self, other):
        """ Return whether the block is equivalent to the other given or not. """
        return Block.equivalent(self, other) and self.number_arguments == other.number_arguments

    def evaluate(self, values, **kwargs):
        """ Pack values into a sequence. """
        return [[values[var] for var in self.inputs]]

    def _to_script(self, _) -> ToScriptElement:
        """ Write block config into a chunk of script. """
        script = f"Sequence(number_arguments={len(self.inputs)}, {self.base_script()})"
        return ToScriptElement(declaration=script, imports=[self.full_classname])


class Concatenate(Block):
    """
    Concatenate the n inputs.

    :param number_arguments: Number of input to, concatenate.
    :param name: Name of the block.
    :param position: Position of the block in canvas.
    """

    def __init__(self, number_arguments: int = 2, name: str = '', position: Tuple[float, float] = None):
        self.number_arguments = number_arguments
        inputs = [Variable(name=f"Sequence element {i}") for i in range(self.number_arguments)]
        outputs = [TypedVariable(type_=List[T], name='sequence')]
        Block.__init__(self, inputs, outputs, name=name, position=position)

    def equivalent_hash(self):
        """ Custom hash function. Related to 'equivalent' method. """
        return self.number_arguments

    def equivalent(self, other):
        """ Return whether the block is equivalent to the other given or not. """
        return Block.equivalent(self, other) and self.number_arguments == other.number_arguments

    def evaluate(self, values: Dict[Variable, Any], **kwargs):
        """ Concatenate elements that are in values. """
        list_values = list(values.values())
        return [concatenate(list_values)]

    def _to_script(self, _) -> ToScriptElement:
        """ Write block config into a chunk of script. """
        script = f"Concatenate(number_arguments={len(self.inputs)}, {self.base_script()})"
        return ToScriptElement(declaration=script, imports=[self.full_classname])


class WorkflowBlock(Block):
    """
    Wrapper around workflow to put it in a block of another workflow.

    Even if a workflow is a block, it can't be used directly as it has a different behavior
    than a Block in eq and hash which is problematic to handle in dictionaries for example.

    :param workflow: The WorkflowBlock's workflow
    :param name: Name of the block.
    :param position: Position of the block in canvas.
    """

    def __init__(self, workflow: Workflow, name: str = '', position: Tuple[float, float] = None):
        self.workflow = workflow
        # TODO: configuring port internal connections
        self.input_connections = None
        self.output_connections = None
        inputs = []
        for variable in self.workflow.inputs:
            input_ = variable.copy()
            input_.name = f"{name} - {variable.name}"
            inputs.append(input_)

        outputs = [self.workflow.output.copy()]
        Block.__init__(self, inputs, outputs, name=name, position=position)

    def equivalent_hash(self):
        """ Custom hash function. Related to 'equivalent' method. """
        return hash(self.workflow)

    def equivalent(self, other):
        """ Return whether the block is equivalent to the other given or not. """
        if not Block.equivalent(self, other):
            return False
        return self.workflow == other.workflow

    def evaluate(self, values, **kwargs):
        """ Format sub workflow arguments and run it. """
        arguments = {self.inputs.index(input_): v for input_, v in values.items()}
        workflow_run = self.workflow.run(arguments)
        return [workflow_run.output_value]

    def package_mix(self):
        """ Recursively add block contribution to workflow's package_mix. """
        return self.workflow.package_mix()

    def _docstring(self):
        """ Recursively get docstring of sub workflow. """
        workflow_docstrings = self.workflow._docstring()
        docstring = {}
        for block_docstring in workflow_docstrings:
            for input_ in self.workflow.inputs:
                if block_docstring and input_ in block_docstring:
                    docstring[input_] = block_docstring[input_]
        return docstring

    def _to_script(self, prefix: str) -> ToScriptElement:
        """ Write block config into a chunk of script. """
        prefix = f'{prefix}sub_'
        workflow_script = self.workflow._to_script(prefix)
        script_workflow = f"\n# --- Subworkflow --- \n" \
            f"{workflow_script.declaration}" \
            f"# --- End Subworkflow --- \n"

        script = f"WorkflowBlock(workflow={prefix}workflow, {self.base_script()})"

        imports = workflow_script.imports + [self.full_classname]
        return ToScriptElement(declaration=script, before_declaration=script_workflow, imports=imports)


class ForEach(Block):
    """
    A block to iterate on an input and perform an parallel for (iterations are not dependent).

    :param workflow_block: The WorkflowBlock on which iterate.
    :param iter_input_index: Index of iterable input in worklow_block.inputs
    :param name: Name of the block.
    :param position: Position of the block in canvas.
    """

    def __init__(self, workflow_block: 'WorkflowBlock', iter_input_index: int, name: str = '',
                 position: Tuple[float, float] = None):
        self.workflow_block = workflow_block
        self.iter_input_index = iter_input_index
        self.iter_input = self.workflow_block.inputs[iter_input_index]
        inputs = []
        for i, workflow_input in enumerate(self.workflow_block.inputs):
            if i == iter_input_index:
                variable_name = 'Iterable input: ' + workflow_input.name
                inputs.append(Variable(name=variable_name))
            else:
                input_ = workflow_input.copy()
                input_.name = 'binding ' + input_.name
                inputs.append(input_)
        output_variable = Variable(name='Foreach output')
        self.output_connections = None  # TODO: configuring port internal connections
        self.input_connections = None
        Block.__init__(self, inputs, [output_variable], name=name, position=position)

    def equivalent_hash(self):
        """ Custom hash function. Related to 'equivalent' method. """
        wb_hash = int(self.workflow_block.equivalent_hash() % 10e5)
        return wb_hash + self.iter_input_index

    def equivalent(self, other):
        """ Return whether the block is equivalent to the other given or not. """
        input_eq = self.iter_input_index == other.iter_input_index
        wb_eq = self.workflow_block.equivalent(other.workflow_block)
        return Block.equivalent(self, other) and wb_eq and input_eq

    def evaluate(self, values, **kwargs):
        """ Loop on input list and run sub workflow on each. """
        values_workflow = {var2: values[var1] for var1, var2 in zip(self.inputs, self.workflow_block.inputs)}
        output_values = []
        for value in values_workflow[self.iter_input]:
            values_workflow[self.iter_input] = value
            output = self.workflow_block.evaluate(values_workflow)[0]
            output_values.append(output)
        return [output_values]

    def _docstring(self):
        """ Recursively get docstring of sub-workflow. """
        wb_docstring = self.workflow_block._docstring()
        block_docstring = {}
        for input_, workflow_input in zip(self.inputs, self.workflow_block.workflow.inputs):
            block_docstring[input_] = wb_docstring[workflow_input]
        return block_docstring

    def _to_script(self, prefix: str) -> ToScriptElement:
        """ Write block config into a chunk of script. """
        wfblock_script_elements = self.workflow_block._to_script(prefix)
        wfblock_script = f"{wfblock_script_elements.before_declaration}\n" \
                         f"wfblock = {wfblock_script_elements.declaration}"
        foreach_script = f"ForEach(workflow_block=wfblock, iter_input_index={self.iter_input_index}, " \
                         f"{self.base_script()})"

        imports = wfblock_script_elements.imports + [self.full_classname]
        return ToScriptElement(declaration=foreach_script, before_declaration=wfblock_script, imports=imports)


class Unpacker(Block):
    """
    DeMUX block.

    :param indices: List of indices to extract.
    :param name: Name of the block.
    :param position: Position of the block in canvas.
    """

    def __init__(self, indices: List[int], name: str = '', position: Tuple[float, float] = None):
        self.indices = indices
        outputs = [Variable(name=f"output_{i}") for i in indices]
        Block.__init__(self, inputs=[Variable(name="input_sequence")], outputs=outputs, name=name, position=position)

    def equivalent(self, other):
        """ Return whether the block is equivalent to the other given or not. """
        return Block.equivalent(self, other) and self.indices == other.indices

    def equivalent_hash(self):
        """ Custom hash function. Related to 'equivalent' method. """
        return len(self.indices)

    def evaluate(self, values, **kwargs):
        """ Unpack input list elements into n outputs. """
        return [values[self.inputs[0]][i] for i in self.indices]

    def _to_script(self, _) -> ToScriptElement:
        """ Write block config into a chunk of script. """
        script = f"Unpacker(indices={self.indices}, {self.base_script()})"
        return ToScriptElement(declaration=script, imports=[self.full_classname])


class Flatten(Block):
    """
    A block to extract the first element of a list and flatten it.

    :param name: Name of the block.
    :param position: Position of the block in canvas.
    """

    def __init__(self, name: str = '', position: Tuple[float, float] = None):
        inputs = [Variable(name='input_sequence')]
        outputs = [Variable(name='flatten_sequence')]
        Block.__init__(self, inputs, outputs, name=name, position=position)

    def equivalent_hash(self):
        """ Custom hash function. Related to 'equivalent' method. """
        return 1

    def evaluate(self, values, **kwargs):
        """ Extract the first element of a list and flatten it. """
        output = []
        for value in values[self.inputs[0]]:
            output.extend(value)
        return [output]

    def _to_script(self, _) -> ToScriptElement:
        """ Write block config into a chunk of script. """
        script = f"Flatten({self.base_script()})"
        return ToScriptElement(declaration=script, imports=[self.full_classname])


class Product(Block):
    """ A block to generate the product combinations. """

    def __init__(self, number_list: int, name: str = '', position: Tuple[float, float] = None):
        self.number_list = number_list
        inputs = [Variable(name='list_product_' + str(i)) for i in range(self.number_list)]
        output_variable = Variable(name='Product output')
        Block.__init__(self, inputs, [output_variable], name=name, position=position)

    def equivalent_hash(self):
        """ Custom hash function. Related to 'equivalent' method. """
        return self.number_list

    def equivalent(self, other):
        """ Return whether the block is equivalent to the other given or not. """
        return Block.equivalent(self, other) and self.number_list == other.number_list

    def evaluate(self, values, **kwargs):
        """ Compute the block: use itertools.product. """
        list_product = [values[var] for var in self.inputs]
        output_value = list(itertools.product(*list_product))
        return [output_value]

    def _to_script(self, _) -> ToScriptElement:
        """ Write block config into a chunk of script. """
        script = f"Product(number_list={self.number_list}, {self.base_script()})"
        return ToScriptElement(declaration=script, imports=[self.full_classname])


class Filter(Block):
    """
    A Block to filter some data according to some filters.

    :param filters: A list of DessiaFilters, each corresponding to a value to filter.
    :param name: Name of the block.
    :param position: Position of the block in canvas.
    """

    def __init__(self, filters: List[DessiaFilter], logical_operator: str = "and", name: str = '',
                 position: Tuple[float, float] = None):
        self.filters = filters
        self.logical_operator = logical_operator
        inputs = [Variable(name='input_list')]
        outputs = [Variable(name='output_list')]
        Block.__init__(self, inputs, outputs, name=name, position=position)

    def equivalent(self, other):
        """ Return whether the block is equivalent to the other given or not. """
        return Block.equivalent(self, other) and self.filters == other.filters

    def equivalent_hash(self):
        """ Custom hash function. Related to 'equivalent' method. """
        hashes = [hash(f) for f in self.filters]
        return int(sum(hashes) % 10e5)

    def evaluate(self, values, **kwargs):
        """ Apply given filters to input list. """
        filters_list = FiltersList(self.filters, self.logical_operator)
        return [filters_list.apply(values[self.inputs[0]])]

    def _to_script(self, _) -> ToScriptElement:
        """ Write block config into a chunk of script. """
        filter_variables = [f"DessiaFilter("
                            f"attribute='{f.attribute}', comparison_operator='{f.comparison_operator}', "
                            f"bound={f.bound}, name='{f.name}')" for f in self.filters]
        filters = '[' + ",".join(filter_variables) + ']'
        script = f"Filter(filters={filters}, logical_operator='{self.logical_operator}', {self.base_script()})"

        imports = [DessiaFilter("", "", 0).full_classname, self.full_classname]
        return ToScriptElement(declaration=script, imports=imports)


class Display(Block):
    """ Abstract block class for display behaviors. """

    _displayable_input = 0
    _non_editable_attributes = ["inputs"]
    _type = None
    serialize = False

    def __init__(self, inputs: List[Variable], load_by_default: bool = False, name: str = "",
                 selector: Optional[ViewType] = None, position: Tuple[float, float] = None):
        output = TypedVariable(type_=DisplayObject, name="Display Object")
        Block.__init__(self, inputs=inputs, outputs=[output], name=name, position=position)

        self.load_by_default = load_by_default
        self.selector = selector

    @property
    def type_(self) -> str:
        """ Get display's type_. """
        if self._type:
            return self._type
        raise NotImplementedError(f"type_ attribute is not implemented for block of type '{type(self)}'")

    def _display_settings(self, block_index: int, reference_path: str = "#") -> DisplaySetting:
        """ Compute block's display settings. """
        arguments = {"block_index": block_index, "reference_path": reference_path}
        return DisplaySetting(selector=self.selector.name, type_=self.type_, method="block_display",
                              serialize_data=self.serialize, arguments=arguments,
                              load_by_default=self.load_by_default)

    def evaluate(self, values, **kwargs):
        """ Run method defined by selector's display_setting and compute corresponding DisplayObject. """
        object_ = values[self.inputs[0]]
        settings = object_._display_settings_from_selector(self.selector.name)
        method = settings.method
        if "progress_callback" in kwargs:
            # User methods do not necessarily implement progress callback
            del kwargs["progress_callback"]
        try:
            return [attrmethod_getter(object_, method)(**kwargs)]
        except TypeError as exc:
            arguments = list(kwargs.keys())
            warnings.warn(f"Workflow : method '{method}' was called without generic arguments "
                          f"('{', '.join(arguments)}') because one of them is not set in method's signature.\n\n "
                          f"Original exception : \n{repr(exc)}")
            # Cover cases where kwargs do not correspond to method signature (missing reference_path, for ex)
            return [attrmethod_getter(object_, method)()]

    def _to_script(self, _) -> ToScriptElement:
        """ Write block config into a chunk of script. """
        script = f"{self.__class__.__name__}(name='{self.name}')"
        return ToScriptElement(declaration=script, imports=[self.full_classname])


class DeprecatedMultiPlot(Display):
    """
    Generate a Multi plot which axes will be the given attributes.

    :param attributes: A List of all attributes that will be shown on axes in the ParallelPlot window.
        Can be deep attributes with the '/' separator.
    :param name: Name of the block.
    :param position: Position of the block in canvas.
    """

    type_ = "plot_data"

    def __init__(self, attributes: List[str], load_by_default: bool = True,
                 name: str = "", position: Tuple[float, float] = None):
        self.attributes = attributes
        Display.__init__(self, inputs=[TypedVariable(List[DessiaObject])], load_by_default=load_by_default,
                         name=name, position=position)
        self.inputs[0].name = "Input List"
        self.serialize = True

    def equivalent(self, other):
        """ Return whether if the block is equivalent to the other given. """
        same_attributes = self.attributes == other.attributes
        return Block.equivalent(self, other) and same_attributes

    def equivalent_hash(self):
        """ Custom hash function. Related to 'equivalent' method. """
        return sum(len(a) for a in self.attributes)

    def evaluate(self, values, **kwargs):
        """ Create MultiPlot from block configuration. Handle reference path. """
        reference_path = kwargs.get("reference_path", "#")
        import plot_data
        objects = values[self.inputs[self._displayable_input]]
        samples = [plot_data.Sample(values={a: get_in_object_from_path(o, a) for a in self.attributes},
                                    reference_path=f"{reference_path}/{i}", name=f"Sample {i}")
                   for i, o in enumerate(objects)]
        samples2d = [plot_data.Sample(values={a: get_in_object_from_path(o, a) for a in self.attributes[:2]},
                                      reference_path=f"{reference_path}/{i}", name=f"Sample {i}")
                     for i, o in enumerate(objects)]
        tooltip = plot_data.Tooltip(name='Tooltip', attributes=self.attributes)

        scatterplot = plot_data.Scatter(tooltip=tooltip, x_variable=self.attributes[0], y_variable=self.attributes[1],
                                        elements=samples2d, name='Scatter Plot')

        parallelplot = plot_data.ParallelPlot(disposition='horizontal', axes=self.attributes,
                                              rgbs=[(192, 11, 11), (14, 192, 11), (11, 11, 192)], elements=samples)
        plots = [scatterplot, parallelplot]
        sizes = [plot_data.Window(width=560, height=300), plot_data.Window(width=560, height=300)]
        multiplot = plot_data.MultiplePlots(elements=samples, plots=plots, sizes=sizes,
                                            coords=[(0, 0), (0, 300)], name='Results plot')
        return [multiplot.to_dict()]

    def _to_script(self, _) -> ToScriptElement:
        """ Write block config into a chunk of script. """
        script = f"MultiPlot(attributes={self.attributes}, {self.base_script()})"
        return ToScriptElement(declaration=script, imports=[self.full_classname])


class MultiPlot(Display):
    """
    Generate a Multiplot which axes will be the given attributes. Will show a Scatter and a Parallel Plot.

    :param selector_name: Name of the selector to be displayed in object page. Must be unique throughout workflow.
    :param attributes: A List of all attributes that will be shown on axes in the ParallelPlot window.
        Can be deep attributes with the '/' separator.
    :param name: Name of the block.
    :param position: Position of the block in canvas.
    """

    _type = "plot_data"
    serialize = True

    def __init__(self, selector_name: str, attributes: List[str], load_by_default: bool = True,
                 name: str = "", position: Tuple[float, float] = None):
        self.attributes = attributes
        Display.__init__(self, inputs=[TypedVariable(List[DessiaObject])], load_by_default=load_by_default,
                         name=name, selector=PlotDataType(class_=DessiaObject, name=selector_name), position=position)
        self.inputs[0].name = "Input List"

    def __deepcopy__(self, memo=None):
        return MultiPlot(selector_name=self.selector.name, attributes=self.attributes,
                         load_by_default=self.load_by_default, name=self.name, position=self.position)

    def equivalent(self, other):
        """ Return whether if the block is equivalent to the other given. """
        same_attributes = self.attributes == other.attributes
        return Block.equivalent(self, other) and same_attributes

    def equivalent_hash(self):
        """ Custom hash function. Related to 'equivalent' method. """
        return sum(len(a) for a in self.attributes)

    def evaluate(self, values, **kwargs):
        """ Create MultiPlot from block configuration. Handle reference path. """
        reference_path = kwargs.get("reference_path", "#")
        import plot_data
        objects = values[self.inputs[self._displayable_input]]
        samples = [plot_data.Sample(values={a: get_in_object_from_path(o, a) for a in self.attributes},
                                    reference_path=f"{reference_path}/{i}", name=f"Sample {i}")
                   for i, o in enumerate(objects)]
        samples2d = [plot_data.Sample(values={a: get_in_object_from_path(o, a) for a in self.attributes[:2]},
                                      reference_path=f"{reference_path}/{i}", name=f"Sample {i}")
                     for i, o in enumerate(objects)]
        tooltip = plot_data.Tooltip(name='Tooltip', attributes=self.attributes)

        scatterplot = plot_data.Scatter(tooltip=tooltip, x_variable=self.attributes[0], y_variable=self.attributes[1],
                                        elements=samples2d, name='Scatter Plot')

        parallelplot = plot_data.ParallelPlot(disposition='horizontal', axes=self.attributes,
                                              rgbs=[(192, 11, 11), (14, 192, 11), (11, 11, 192)], elements=samples)
        plots = [scatterplot, parallelplot]
        sizes = [plot_data.Window(width=560, height=300), plot_data.Window(width=560, height=300)]
        multiplot = plot_data.MultiplePlots(elements=samples, plots=plots, sizes=sizes,
                                            coords=[(0, 0), (0, 300)], name='Results plot')
        return [multiplot.to_dict()]

    def _to_script(self, _) -> ToScriptElement:
        """ Write block config into a chunk of script. """
        script = f"MultiPlot(attributes={self.attributes}, {self.base_script()})"
        return ToScriptElement(declaration=script, imports=[self.full_classname])

    def to_dict(self, use_pointers: bool = True, memo=None, path: str = '#',
                id_method=True, id_memo=None) -> JsonSerializable:
        """ Overwrite to_dict method in order to handle difference of behaviors about selector. """
        dict_ = super().to_dict(use_pointers=use_pointers, memo=memo, path=path, id_method=id_method, id_memo=id_memo)
        dict_.update({"selector_name": self.selector.name, "attributes": self.attributes, "name": self.name,
                      "load_by_default": self.load_by_default, "position": self.position})
        return dict_

    @classmethod
    def dict_to_object(cls, dict_: JsonSerializable, force_generic: bool = False, global_dict=None,
                       pointers_memo: Dict[str, Any] = None, path: str = '#'):
        """ Backward compatibility for old versions of Display blocks. """
        selector_name = dict_.get("selector_name", None)
        selector = dict_.get("selector", None)
        if selector is None and selector_name is None:
            # Backward compatibility < 0.14.0
            load_by_default = dict_.get("load_by_default", False)
            return DeprecatedMultiPlot(attributes=dict_["attributes"], name=dict_["name"],
                                       load_by_default=load_by_default, position=dict_["position"])
        if selector is not None and selector_name is None:
            if isinstance(selector, str):
                selector_name = selector
            else:
                # Backward compatibility 0.14.0 < v < 0.14.1
                selector_name = selector["name"]
        return MultiPlot(selector_name=selector_name, attributes=dict_["attributes"], name=dict_["name"],
                         load_by_default=dict_["load_by_default"], position=dict_["position"])


class DeprecatedCadView(Display):
    """
    Deprecated version of CadView block.

    Steps to upgrade to new version (CadView) :
    - Remove type_ argument from your __init__ call
    - Argument 'selector' is now the first in the call and doesn't have default value anymore.
    - Argument 'selector' is of type CadViewType instead of str
    """

    _type = "babylon_data"

    def __init__(self, name: str = "", load_by_default: bool = False, selector: str = "cad",
                 position: Tuple[float, float] = None):
        warnings.warn("This version of CadView Block is deprecated and should not be used anymore."
                      "Please upgrade to CadView new version, instead. (see docstrings)", DeprecationWarning)
        input_ = TypedVariable(DessiaObject, name="Model to display")
        Display.__init__(self, inputs=[input_], load_by_default=load_by_default, selector=selector,
                         name=name, position=position)


class CadView(Display):
    """ Generate a DisplayObject that is displayable in 3D Viewer features (BabylonJS, ...). """

    _type = "babylon_data"

    def __init__(self, selector: CadViewType[Type], name: str = "", load_by_default: bool = False,
                 position: Tuple[float, float] = None):
        if isinstance(selector, str):
            raise TypeError("Argument 'selector' should be of type 'CadViewType' and not 'str',"
                            " which is deprecated. See upgrading guide if needed.")
        input_ = TypedVariable(DessiaObject, name="Model to display")
        Display.__init__(self, inputs=[input_], load_by_default=load_by_default, selector=selector,
                         name=name, position=position)

    @classmethod
    def dict_to_object(cls, dict_: JsonSerializable, force_generic: bool = False, global_dict=None,
                       pointers_memo: Dict[str, Any] = None, path: str = '#'):
        """ Backward compatibility for old versions of Display blocks. """
        selector = dict_.get("selector", "cad")
        if isinstance(selector, str):
            load_by_default = dict_.get("load_by_default", False)
            return DeprecatedCadView(name=dict_["name"], load_by_default=load_by_default, selector=selector,
                                     position=dict_["position"])
        selector = CadViewType.dict_to_object(selector)
        return CadView(selector=selector, name=dict_["name"], load_by_default=dict_["load_by_default"],
                       position=dict_["position"])


class DeprecatedMarkdown(Display):
    """
    Deprecated version of Markdown block.

    Steps to upgrade to new version (Markdown) :
    - Remove type_ argument from your __init__ call
    - Argument 'selector' is now the first in the call and doesn't have default value anymore.
    - Argument 'selector' is of type MarkdownSelector instead of str.
    """

    _type = "markdown"

    def __init__(self, name: str = "", load_by_default: bool = False, selector: str = "markdown",
                 position: Tuple[float, float] = None):
        warnings.warn("This version of 'Markdown' Block is deprecated and should not be used anymore."
                      "Please upgrade to 'Markdown' new version, instead. (see docstrings)", DeprecationWarning)
        input_ = TypedVariable(DessiaObject, name="Model to display")
        Display.__init__(self, inputs=[input_], load_by_default=load_by_default, name=name,
                         selector=selector, position=position)


class Markdown(Display):
    """ Generate a Markdown representation of an object. """

    _type = "markdown"

    def __init__(self, selector: MarkdownType[Type], name: str = "", load_by_default: bool = False,
                 position: Tuple[float, float] = None):
        if isinstance(selector, str):
            raise TypeError("Argument 'selector' should be of type 'MarkdownType' and not 'str',"
                            " which is deprecated. See upgrading guide if needed.")
        input_ = TypedVariable(DessiaObject, name="Model to display")
        Display.__init__(self, inputs=[input_], load_by_default=load_by_default, selector=selector,
                         name=name, position=position)

    @classmethod
    def dict_to_object(cls, dict_: JsonSerializable, force_generic: bool = False, global_dict=None,
                       pointers_memo: Dict[str, Any] = None, path: str = '#'):
        """ Backward compatibility for old versions of Display blocks. """
        selector = dict_.get("selector", "markdown")
        if isinstance(selector, str):
            load_by_default = dict_.get("load_by_default", False)
            return DeprecatedMarkdown(name=dict_["name"], load_by_default=load_by_default, selector=selector,
                                      position=dict_["position"])
        selector = MarkdownType.dict_to_object(selector)
        return Markdown(selector=selector, name=dict_["name"], load_by_default=dict_["load_by_default"],
                        position=dict_["position"])


class DeprecatedPlotData(Display):
    """
    Deprecated version of PlotData block.

    Steps to upgrade to new version (PlotData) :
    - Remove type_ argument from your __init__ call
    - Argument 'selector' is now the first in the call and doesn't have default value anymore.
    - Argument 'selector' is of type PlotDataSelector instead of str.
    """

    _type = "plot_data"
    serialize = True

    def __init__(self, name: str = '', load_by_default: bool = False, selector: str = "plot_data",
                 position: Tuple[float, float] = None):
        warnings.warn("This version of 'PlotData' Block is deprecated and should not be used anymore."
                      "Please upgrade to 'PlotData' new version, instead. (see docstrings)", DeprecationWarning)
        input_ = TypedVariable(DessiaObject, name="Model to display")
        Display.__init__(self, inputs=[input_], load_by_default=load_by_default, name=name,
                         selector=selector, position=position)


class PlotData(Display):
    """ Generate a PlotData representation of an object. """

    _type = "plot_data"
    serialize = True

    def __init__(self, selector: PlotDataType[Type], name: str = "", load_by_default: bool = False,
                 position: Tuple[float, float] = None):
        if isinstance(selector, str):
            raise TypeError("Argument 'selector' should be of type 'PlotDataType' and not 'str',"
                            " which is deprecated. See upgrading guide if needed.")
        input_ = TypedVariable(DessiaObject, name="Model to display")
        Display.__init__(self, inputs=[input_], load_by_default=load_by_default, selector=selector,
                         name=name, position=position)

    @classmethod
    def dict_to_object(cls, dict_: JsonSerializable, force_generic: bool = False, global_dict=None,
                       pointers_memo: Dict[str, Any] = None, path: str = '#'):
        """ Backward compatibility for old versions of Display blocks. """
        selector = dict_.get("selector", "plot_data")
        if isinstance(selector, str):
            load_by_default = dict_.get("load_by_default", False)
            return DeprecatedPlotData(name=dict_["name"], load_by_default=load_by_default, selector=selector,
                                      position=dict_["position"])
        selector = PlotDataType.dict_to_object(selector)
        return PlotData(selector=selector, name=dict_["name"], load_by_default=dict_["load_by_default"],
                        position=dict_["position"])


class ModelAttribute(Block):
    """
    Fetch attribute of given object during workflow execution.

    :param attribute_name: The name of the attribute to select.
    :param name: Name of the block.
    :param position: Position of the block in canvas.
    """

    def __init__(self, attribute_name: str, name: str = '', position: Tuple[float, float] = None):
        self.attribute_name = attribute_name
        inputs = [Variable(name='Model')]
        outputs = [Variable(name='Model attribute')]
        Block.__init__(self, inputs, outputs, name=name, position=position)

    def equivalent_hash(self):
        """ Custom hash function. Related to 'equivalent' method. """
        return len(self.attribute_name)

    def equivalent(self, other):
        """ Return whether the block is equivalent to the other given or not. """
        return Block.equivalent(self, other) and self.attribute_name == other.attribute_name

    def evaluate(self, values, **kwargs):
        """ Get input object's deep attribute. """
        return [get_in_object_from_path(values[self.inputs[0]], f'#/{self.attribute_name}')]

    def _to_script(self, _) -> ToScriptElement:
        """ Write block config into a chunk of script. """
        script = f"ModelAttribute(attribute_name='{self.attribute_name}', {self.base_script()})"
        return ToScriptElement(declaration=script, imports=[self.full_classname])

   
class GetModelAttribute(Block):
    """
    Fetch attribute of given object during workflow execution.

    :param attribute_type: AttributeType variable that contain the model and the name of the attribute to select.
    :param name: Name of the block.
    :param position: Position of the block in canvas.
    """

    def __init__(self, attribute_type: AttributeType[Type], name: str = '', position: Tuple[float, float] = None):
        self.attribute_type = attribute_type
        parameters = inspect.signature(self.attribute_type.class_).parameters
        inputs = [TypedVariable(type_=self.attribute_type.class_, name='Model')]
        type_ = get_attribute_type(self.attribute_type.name, parameters)
        if type_:
            outputs = [TypedVariable(type_=type_, name='Model attribute')]  
        else:
            outputs = [Variable(name='Model attribute')]
        Block.__init__(self, inputs, outputs, name=name, position=position)

    def equivalent_hash(self):
        """ Custom hash function. Related to 'equivalent' method. """
        classname = self.attribute_type.class_.__name__
        return len(classname) + 7 * len(self.attribute_type.name)

    def equivalent(self, other):
        """ Return whether the block is equivalent to the other given or not. """
        classname = self.attribute_type.class_.__name__
        other_classname = other.attribute_type.class_.__name__
        same_model = classname == other_classname
        same_method = self.attribute_type.name == other.attribute_type.name
        return Block.equivalent(self, other) and same_model and same_method

    def evaluate(self, values, **kwargs):
        """ Get input object's deep attribute. """
        return [get_in_object_from_path(values[self.inputs[0]], f'#/{self.attribute_type.name}')]

    def _to_script(self, _) -> ToScriptElement:
        """ Write block config into a chunk of script. """
        script = f"GetModelAttribute(attribute_type=AttributeType(" \
                 f"{self.attribute_type.class_.__name__}, name=\"{self.attribute_type.name}\")" \
                 f", {self.base_script()})"
        imports = [full_classname(object_=self.attribute_type, compute_for='instance'),
                   full_classname(object_=self.attribute_type.class_, compute_for='class'),
                   self.full_classname]
        return ToScriptElement(declaration=script, imports=imports)

    @classmethod
    def dict_to_object(cls, dict_: JsonSerializable, force_generic: bool = False, global_dict=None,
                       pointers_memo: Dict[str, Any] = None, path: str = '#'):
        """ Backward compatibility for old versions of blocks. """
        # Backward compatibility dessia_common < 0.14.0
        if "object_class" not in dict_["attribute_type"]:
            dict_["attribute_type"]["object_class"] = "dessia_common.typings.AttributeType"
        return super().dict_to_object(dict_=dict_, force_generic=True, global_dict=global_dict,
                                      pointers_memo=pointers_memo, path=path)


class SetModelAttribute(Block):
    """
    Block to set an attribute value in a workflow.

    :param attribute_type: AttributeType variable that contain the model and the name of the attribute to select.
    :param name: Name of the block.
    :param position: Position of the block in canvas.
    """

    def __init__(self, attribute_type: AttributeType[Type], name: str = '', position: Tuple[float, float] = None):
        self.attribute_type = attribute_type
        parameters = inspect.signature(self.attribute_type.class_).parameters
        type_ = get_attribute_type(self.attribute_type.name, parameters)
        inputs = [TypedVariable(type_=self.attribute_type.class_, name='Model')]
        if type_:
            inputs.append(TypedVariable(type_=type_, name=f'Value to insert for attribute {self.attribute_type.name}'))
        else:
            inputs.append(Variable(name=f'Value to insert for attribute {self.attribute_type.name}'))
        outputs = [TypedVariable(type_=self.attribute_type.class_,
                                 name=f'Model with changed attribute {self.attribute_type.name}')]
        Block.__init__(self, inputs, outputs, name=name, position=position)

    def equivalent_hash(self):
        """ Custom hash function. Related to 'equivalent' method. """
        return 3 + len(self.attribute_type.name)

    def equivalent(self, other):
        """ Returns whether the block is equivalent to the other given or not. """
        return Block.equivalent(self, other) and self.attribute_type.name == other.attribute_type.name

    def evaluate(self, values, **kwargs):
        """ Set input object's deep attribute with input value. """
        model = values[self.inputs[0]]
        setattr(model, self.attribute_type.name, values[self.inputs[1]])
        return [model]

    def _to_script(self, _) -> ToScriptElement:
        """ Write block config into a chunk of script. """
        script = f"SetModelAttribute(attribute_type=AttributeType(" \
                 f"{self.attribute_type.class_.__name__}, name=\"{self.attribute_type.name}\")" \
                 f", {self.base_script()})"
        return ToScriptElement(declaration=script, imports=[self.full_classname])

    @classmethod
    def dict_to_object(cls, dict_: JsonSerializable, force_generic: bool = False, global_dict=None,
                       pointers_memo: Dict[str, Any] = None, path: str = '#'):
        """ Backward compatibility for old versions of blocks. """
        # Backward compatibility dessia_common < 0.14.0
        if "object_class" not in dict_["attribute_type"]:
            dict_["attribute_type"]["object_class"] = "dessia_common.typings.AttributeType"
        return super().dict_to_object(dict_=dict_, force_generic=True, global_dict=global_dict,
                                      pointers_memo=pointers_memo, path=path)


class Sum(Block):
    """
    Sum the n inputs.

    :param number_elements: Number of element to sum
    :param name: Name of the block
    :param position: Position of the block in the workflow
    """

    def __init__(self, number_elements: int = 2, name: str = '', position: Tuple[float, float] = None):
        self.number_elements = number_elements
        inputs = [Variable(name=f"Sum element {i + 1}") for i in range(number_elements)]
        Block.__init__(self, inputs=inputs, outputs=[Variable(name='Sum')], name=name, position=position)

    def equivalent_hash(self):
        """ Custom hash function. Related to 'equivalent' method. """
        return self.number_elements

    def equivalent(self, other):
        """ Returns whether the block is equivalent to the other given or not. """
        return Block.equivalent(self, other) and self.number_elements == other.number_elements

    def evaluate(self, values, **kwargs):
        """
        Sum input values.

        TODO : This cannot work, we are summing a dictionary
        """
        return [sum(values)]

    def _to_script(self, _) -> ToScriptElement:
        """ Write block config into a chunk of script. """
        script = f"Sum(number_elements={self.number_elements}, {self.base_script()})"
        return ToScriptElement(declaration=script, imports=[self.full_classname])


class Substraction(Block):
    """ Block that subtract input values. First is +, second is -. """

    def __init__(self, name: str = '', position: Tuple[float, float] = None):
        Block.__init__(self, [Variable(name='+'), Variable(name='-')], [Variable(name='Substraction')], name=name,
                       position=position)

    def evaluate(self, values, **kwargs):
        """ Subtract input values. """
        return [values[self.inputs[0]] - values[self.inputs[1]]]

    def _to_script(self, _) -> ToScriptElement:
        """ Write block config into a chunk of script. """
        script = f"Substraction({self.base_script()})"
        return ToScriptElement(declaration=script, imports=[self.full_classname])


class ConcatenateStrings(Block):
    """
    Concatenate the n input elements, separate by the separator input, into one string.

    :param number_elements: Number of block inputs
    :param separator: Character used to joins the input elements together
    :param name: Name of the block.
    :param position: Position of the block in canvas.
    """

    def __init__(self, number_elements: int = 2, separator: str = "", name: str = '',
                 position: Tuple[float, float] = None):
        self.number_elements = number_elements
        self.separator = separator
        inputs = [TypedVariableWithDefaultValue(name=f"Substring {i + 1}", type_=str, default_value="")
                  for i in range(number_elements)]
        output = TypedVariable(name="Concatenation", type_=str)
        Block.__init__(self, inputs=inputs, outputs=[output], name=name, position=position)

    def equivalent_hash(self):
        """ Custom hash function. Related to 'equivalent' method. """
        return self.number_elements + hash(self.separator)

    def equivalent(self, other):
        """ Returns whether the block is equivalent to the other given or not. """
        same_number = self.number_elements == other.number_elements
        same_separator = self.separator == other.separator
        return Block.equivalent(self, other) and same_number and same_separator

    def evaluate(self, values, **kwargs):
        """ Concatenate input strings with configured separator. """
        chunks = [values[i] for i in self.inputs]
        return [self.separator.join(chunks)]

    def _to_script(self, _) -> ToScriptElement:
        """ Write block config into a chunk of script. """
        script = f"ConcatenateStrings(number_elements={self.number_elements}, " \
                 f"separator='{self.separator}', " \
                 f"name='{self.name}')"
        return ToScriptElement(declaration=script, imports=[self.full_classname])


class Export(Block):
    """
    Block that enables an export of an object calling its configured method.

    Only Methods that yields streams (and not files) should be used.
    The file generated will be called {filename}.{extension}

    :param method_type: An object that have a class_ input (which is the class of the incoming model)
        and a name (which is the name of the method that will be called).
    :param text: Whether the export is of type text or not
    :param extension: Extension of the resulting file (ex: json or xlsx)
    :param filename: Name of the resulting file without its extension
    :param name: Name of the block.
    """

    def __init__(self, method_type: MethodType[Type], text: bool, extension: str,
                 filename: str = "export", name: str = "", position: Tuple[float, float] = None):
        self.method_type = method_type
        if not filename:
            filename = "export"
        self.filename = filename
        method = method_type.get_method()
        self.extension = extension
        self.text = text

        output = output_from_function(function=method, name="export_output")
        inputs = [TypedVariable(type_=method_type.class_, name="model_to_export"),
                  TypedVariableWithDefaultValue(type_=str, default_value=filename, name="filename")]
        Block.__init__(self, inputs=inputs, outputs=[output], name=name, position=position)

    def evaluate(self, values, **kwargs):
        """ Generate to-be-exported stream from corresponding method. """
        filename = f"{values.pop(self.inputs[-1])}.{self.extension}"
        if self.text:
            stream = StringFile(filename)
        else:
            stream = BinaryFile(filename)
        getattr(values[self.inputs[0]], self.method_type.name)(stream)
        return [stream]

    def _export_format(self, block_index: int) -> ExportFormat:
        """ Compute block's export format. """
        arguments = {"block_index": block_index}
        return ExportFormat(selector=None, extension=self.extension, method_name="export", text=self.text,
                            export_name=self.filename, args=arguments)

    def _to_script(self, _) -> ToScriptElement:
        """ Write block config into a chunk of script. """
        script = f"Export(method_type=MethodType(" \
                 f"{self.method_type.class_.__name__}, '{self.method_type.name}')" \
                 f", filename='{self.filename}', extension='{self.extension}'" \
                 f", text={self.text}, {self.base_script()})"

        imports = [self.full_classname, full_classname(object_=self.method_type, compute_for='instance'),
                   full_classname(object_=self.method_type.class_, compute_for='class')]
        return ToScriptElement(declaration=script, imports=imports)


class Archive(Block):
    """
    A block that takes n inputs and store them in a archive (ZIP,...).

    :param number_exports: The number of files that will be stored in the archive
    :param filename: Name of the resulting archive file without its extension
    :param name: Name of the block.
    """

    def __init__(self, number_exports: int = 1, filename: str = "archive", name: str = "",
                 position: Tuple[float, float] = None):
        self.number_exports = number_exports
        self.filename = filename
        self.extension = "zip"
        self.text = False
        inputs = [Variable(name="export_" + str(i)) for i in range(number_exports)]
        inputs.append(TypedVariableWithDefaultValue(type_=str, default_value=filename, name="filename"))
        Block.__init__(self, inputs=inputs, outputs=[Variable(name="zip archive")], name=name, position=position)

    def to_dict(self, use_pointers: bool = True, memo=None, path: str = '#', id_method=True, id_memo=None):
        """ Serialize the block with custom logic. """
        dict_ = Block.to_dict(self, use_pointers=use_pointers, memo=memo, path=path)
        dict_['number_exports'] = len(self.inputs) - 1   # Filename is also a block input
        dict_["filename"] = self.filename
        return dict_

    @classmethod
    @set_block_variable_names_from_dict
    def dict_to_object(cls, dict_: JsonSerializable, force_generic: bool = False,
                       global_dict=None, pointers_memo: Dict[str, Any] = None, path: str = '#'):
        """ Custom dict_to_object method. """
        return cls(number_exports=dict_["number_exports"], filename=dict_["filename"],
                   name=dict_['name'], position=dict_.get('position'))

    def evaluate(self, values, **kwargs):
        """ Generate archive stream for input streams. """
        name_input = self.inputs[-1]
        archive_name = f"{values.pop(name_input)}.{self.extension}"
        archive = BinaryFile(archive_name)
        with ZipFile(archive, 'w') as zip_archive:
            for input_ in self.inputs[:-1]:  # Filename is last block input
                value = values[input_]
                generate_archive(zip_archive, value)
        return [archive]

    def _export_format(self, block_index: int) -> ExportFormat:
        """ Compute block's export formats. """
        arguments = {"block_index": block_index}
        return ExportFormat(selector=None, extension=self.extension, method_name="export", text=self.text,
                            export_name=self.filename, args=arguments)

    def _to_script(self, _) -> ToScriptElement:
        """ Write block config into a chunk of script. """
        script = f"Archive(number_exports={self.number_exports}, filename='{self.filename}', {self.base_script()})"
        return ToScriptElement(declaration=script, imports=[self.full_classname])


def get_attribute_type(attribute_name: str, parameters):
    """ Get type of attribute name of class."""
    parameter = parameters.get(attribute_name)
    if not parameter:
        return None
    if not hasattr(parameter, "annotation"):
        return parameter
    if parameter.annotation == inspect.Parameter.empty:
        return None 
    return parameter.annotation



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

"""

import json
from dessia_common.core import DessiaObject

from dessia_common import REF_MARKER
from dessia_common.models.workflows import simulation_workflow
from dessia_common.models.power_test import components1, component_connections1, usage1
from dessia_common.serialization import serialize

simulation_workflow.to_dict(use_pointers=False)
input_values = {0: components1,
                1: component_connections1,
                3: usage1}

workflow_run = simulation_workflow.run(input_values)

print(workflow_run.log)

arguments = {str(k): serialize(v) for k, v in input_values.items()}
arguments = simulation_workflow.dict_to_arguments(arguments, 'run')

workflow_run2 = DessiaObject.dict_to_object(workflow_run.to_dict())
workflow_run._check_platform()
workflow_run.to_dict(use_pointers=False)

manual_run = simulation_workflow.start_run({0: components1, 1: component_connections1})

# print(manual_run)

evaluated_block = manual_run.evaluate_next_block()
assert evaluated_block is not None
# print(evaluated_block)

evaluated = manual_run.block_evaluation(1)
# print(evaluated)
assert not evaluated

manual_run.add_input_value(3, usage1)

evaluated_blocks = manual_run.continue_run()
assert(manual_run.progress == 1)
print('progress: ', manual_run.progress)

manual_run._displays()
manual_run.to_dict(use_pointers=False)

manual_run.schema()
manual_run.performance_analysis()

# Testing that there is no pointer when use_pointers=False
d = workflow_run.to_dict(use_pointers=False)
s = json.dumps(d)
if REF_MARKER in s:
    ind_ref = s.index(REF_MARKER)
    print(s[ind_ref - 300:ind_ref + 500])
    raise ValueError('Pointer detected with use_pointers=False')

simulation_workflow.save_script_to_file('_simulation_workflow')

print("script power_simulation.py has passed")



"""
Tests for dessia_common.Dataset class (loadings, check_platform and plots)
"""
import random
import numpy as npy
from dessia_common.models import all_cars_wi_feat
from dessia_common.datatools.dataset import Dataset
from dessia_common.datatools.cluster import ClusteredDataset

# =============================================================================
# TEST PARETO FRONT
# =============================================================================
# Uniform
coord_1 = [random.uniform(0, 0.1) for i in range(1000)]
coord_2 = [random.uniform(0.9e6, 1e6) for i in range(1000)]
costs = npy.array([coord_1, coord_2]).T

pareto_points = Dataset.pareto_indexes(costs)
pareto_frontiers = Dataset.pareto_frontiers(len(costs), costs)

# Uniform
coord_1 = [random.uniform(0, 0.001) for i in range(1000)]
coord_2 = [random.uniform(0, 1e6) for i in range(1000)]
costs = npy.array([coord_1, coord_2]).T

pareto_points = Dataset.pareto_indexes(costs)
pareto_frontiers = Dataset.pareto_frontiers(len(costs), costs)

# Gaussian
coord_1 = [random.gauss(50000, 1) for i in range(1000)]
coord_2 = [random.gauss(10, 1) for i in range(1000)]
costs = npy.array([coord_1, coord_2]).T

pareto_points = Dataset.pareto_indexes(costs)
pareto_frontiers = Dataset.pareto_frontiers(len(costs), costs)

# Cars
all_cars_with_features = Dataset(all_cars_wi_feat)
costs_attributes = ['weight', 'mpg']
pareto_points = all_cars_with_features.pareto_points(costs_attributes)

costs = [all_cars_with_features.attribute_values('weight'), all_cars_with_features.attribute_values('mpg')]
costs = list(zip(*costs))
pareto_frontiers = Dataset.pareto_frontiers(len(all_cars_wi_feat), costs)

# With transposed costs
transposed_costs = list(zip(*costs))
pareto_frontiers = Dataset.pareto_frontiers(len(all_cars_wi_feat), transposed_costs)
assert(npy.array_equal(pareto_frontiers[0], npy.array([[1613., 34.392097264437695], [5140.0, -297.9392097264438]])))
assert(npy.array_equal(pareto_frontiers[1], npy.array([[0.0, 1928.0], [46.6, 1508.6]])))

categorized_pareto = ClusteredDataset.from_pareto_sheets(all_cars_with_features, costs_attributes, 7)
pareto_plot_data = categorized_pareto.plot_data()
# assert(json.dumps(pareto_plot_data[0].to_dict())[150:200] == ', "Cluster Label": 0}, {"mpg": 0.0, "displacement"')
# assert(json.dumps(pareto_plot_data[1].to_dict())[10500:10550] == 't": 2901.0, "acceleration": 16.0, "Cluster Label":')
# assert(json.dumps(pareto_plot_data[2].to_dict())[50:100] == 'te_names": ["Index of reduced basis vector", "Sing')

costs = all_cars_with_features.matrix
categorized_pareto = ClusteredDataset.from_pareto_sheets(all_cars_with_features, costs_attributes, 1)
pareto_plot_data = categorized_pareto.plot_data()
# assert(json.dumps(pareto_plot_data[0].to_dict())[150:200] == ' "Cluster Label": 0}, {"mpg": 14.0, "displacement"')
# assert(json.dumps(pareto_plot_data[1].to_dict())[10500:10550] == 'acceleration": 8.5, "Cluster Label": 1}, {"mpg": 0')
# assert(json.dumps(pareto_plot_data[2].to_dict())[50:100] == 'te_names": ["Index of reduced basis vector", "Sing')

# Missing tests after coverage report
try:
    ClusteredDataset.from_pareto_sheets(all_cars_with_features, ['sddd'], 1)
except Exception as e:
    assert(e.args[0] == ("sddd not in common_attributes = ['mpg', 'displacement', 'horsepower', 'acceleration', " +
                         "'weight']"))



""" A workflow that has 2D, 3D and MD displays. """

from dessia_common.workflow.core import Workflow, Pipe
from dessia_common.workflow.blocks import ClassMethod, PlotData, Markdown
from dessia_common.forms import StandaloneObject
from dessia_common.typings import ClassMethodType, PlotDataType, MarkdownType

plotdata_selector = PlotDataType(class_=StandaloneObject, name="Scatter Plot")
plotdata = PlotData(selector=plotdata_selector, name="2D", load_by_default=True)

markdown_selector = MarkdownType(class_=StandaloneObject, name="Markdown")
markdown = Markdown(selector=markdown_selector, name="MD", load_by_default=False)

cmt = ClassMethodType(class_=StandaloneObject, name="generate")
cmb = ClassMethod(method_type=cmt, name="Generator")


blocks = [cmb, plotdata, markdown]
pipes = [Pipe(input_variable=cmb.outputs[0], output_variable=plotdata.inputs[0]),
         Pipe(input_variable=cmb.outputs[0], output_variable=markdown.inputs[0])]

workflow = Workflow(blocks=blocks, pipes=pipes, output=cmb.outputs[0])
